from nl4dv.utils import constants, helpers
import re
import nltk
from nltk import word_tokenize
from si_prefix import si_parse
import random

class ConversationGenie:
    def __init__(self, nl4dv_instance):
        self.nl4dv_instance = nl4dv_instance
        self.key_num = 0
        self.most_recent_dialog_id = list()
        self.most_recent_query_id = list()
        self.all_dialogs = dict()
        self.branches = dict()

    def add_dialog(self, nl4dv_output, dialog = None, dialog_id = None, query_id = None):
        if dialog is None:
            if dialog_id is not None:
                raise RuntimeError("Dialog is false but conversation id given")
            if query_id is not None:
                raise RuntimeError("Dialog is false but context id given")
            key_str = str(self.key_num)
            self.all_dialogs[key_str] = list()
            self.all_dialogs[key_str].append(nl4dv_output)
            davo = str(self.key_num)
            self.key_num += 1
            self.branches[str(davo)] = 0
            self.most_recent_dialog_id.append(davo)
            self.most_recent_query_id.append(str(0))
            return (davo, str(0))
        else:
            if dialog_id is not None:
                if query_id is not None:
                    if not isinstance(query_id, str):
                        raise RuntimeError("Context id must be of type string")
                    query_id = int(query_id)
                    if query_id == (len(self.all_dialogs[dialog_id]) - 1):
                        self.all_dialogs[dialog_id].append(nl4dv_output)
                        return (dialog_id, str(len(self.all_dialogs[dialog_id]) - 1))
                    else:
                        key_str = str(dialog_id) + "." + str(query_id) + "." + str(self.branches[dialog_id])
                        self.all_dialogs[key_str] = list()
                        self.all_dialogs[key_str].append(self.all_dialogs[dialog_id][query_id])
                        self.all_dialogs[key_str].append(nl4dv_output)
                        davo = key_str
                        self.branches[dialog_id] += 1
                        self.branches[key_str] = 0
                        self.most_recent_dialog_id.append(davo)
                        self.most_recent_query_id.append(str(1))
                        return (davo, "1")
                else:
                    self.all_dialogs[dialog_id].append(nl4dv_output)
                    self.most_recent_dialog_id.append(dialog_id)
                    self.most_recent_query_id.append(str(len(self.all_dialogs[dialog_id]) - 1))
                    return (dialog_id, str(len(self.all_dialogs[dialog_id]) - 1))
            else:
                if query_id is None:
                    key_str = str(self.key_num - 1)

                    self.all_dialogs[key_str].append(nl4dv_output)

                    davo = key_str

                    self.most_recent_dialog_id.append(davo)
                    self.most_recent_query_id.append(str(len(self.all_dialogs[key_str]) - 1))
                    return (davo, str(len(self.all_dialogs[key_str]) - 1))
                else:
                    raise RuntimeError("Dialog is true and context id given but not conversation id")

    def delete_dialogs(self, dialog_id=None, query_id = None):
        # if context id is not None then delete context id and everything after
        if dialog_id is None and query_id is None:
            self.all_dialogs = dict()
            self.branches = dict()
            self.key_num = 0
            return
        elif query_id is None:
            del self.all_dialogs[dialog_id]
            for i in list(self.all_dialogs.keys()):
                if i[:len(dialog_id)] == dialog_id:
                    del self.all_dialogs[dialog_id]
            return
        elif query_id is not None and dialog_id is None:
            raise RuntimeError("Query id given but not Dialog id")
        else:
            if not isinstance(query_id, str):
                raise RuntimeError("Context id must be of type string")
            query_id = int(query_id)
            if dialog_id in self.all_dialogs.keys():
                self.all_dialogs[dialog_id] = self.all_dialogs[dialog_id][:query_id]
            followup_str = str(dialog_id) + "." + str(query_id)
            for i in list(self.all_dialogs.keys()):
                if followup_str == i[:len(followup_str)]:
                    del self.all_dialogs[i]
        return

    def undo(self):
        if not bool(self.all_dialogs):
            return
        recent_dialog_id = self.most_recent_dialog_id[-1]

        recent_query_id = self.most_recent_query_id[-1]

        context_obj = self.all_dialogs[recent_dialog_id][int(recent_query_id)]
        del self.all_dialogs[recent_dialog_id][int(recent_query_id)]
        return context_obj

    def rename_dialog(self, dialog_id, new_name):
        davo = self.all_dialogs[dialog_id]
        self.all_dialogs[new_name] = davo
        del self.all_dialogs[dialog_id]

    def update_query(self, dialog_id, query_id, output):
        if not isinstance(query_id, str):
                raise RuntimeError("Context id must be of type string")
        query_id = int(query_id)
        self.all_dialogs[dialog_id][query_id] = output

    def list_dialogs(self):
        return list(self.all_dialogs.keys())

    def list_all_queries_for_dialog(self, dialog_id):
        return self.all_dialogs[dialog_id]

    def list_all_queries(self):
        return self.all_dialogs

    def get_dialogs(self, dialog_id=None, query_id=None):
        if dialog_id is None and query_id is None:
            return self.all_dialogs
        elif dialog_id is not None and query_id is None:
            return self.all_dialogs[dialog_id]
        elif query_id is not None and dialog_id is None:
            raise RuntimeError("Query id given but not Dialog id")
        else:
            if not isinstance(query_id, str):
                raise RuntimeError("Context id must be of type string")
            query_id = int(query_id)
            return self.all_dialogs[dialog_id][query_id]

    def followup_conversation_attribute(self, extracted_attributes, past_extracted_attributes, attribute_keyword_mapping, past_attribute_keyword_mapping, task_map):
        copy_past = past_extracted_attributes.copy()
        copy_past2 = past_attribute_keyword_mapping.copy()
        for i in extracted_attributes.keys():
            if extracted_attributes[i]['meta']['followup_type'] == 'add':
                copy_past[i] = extracted_attributes[i]
                copy_past2[i] = attribute_keyword_mapping[i]
            elif extracted_attributes[i]['meta']['followup_type'] == 'remove':
                if i in copy_past.keys():
                    del copy_past[i]
                    del copy_past2[i]
                else:
                    helpers.cond_print("Key not found in past query")
                    if 'attribute_domain_value_match' in extracted_attributes[i]['metric']:
                        copy_past[i] = extracted_attributes[i]
                        copy_past2[i] = attribute_keyword_mapping[i]
            elif extracted_attributes[i]['meta']['followup_type'] == 'replace_to_add':
                attr = None
                for j in copy_past.keys():
                    if j in extracted_attributes.keys():
                        if extracted_attributes[j]['meta']['followup_type'] == 'replace_to_remove':
                            attr = j
                if attr in copy_past.keys():
                    del copy_past[attr]
                    del copy_past2[attr]
                else:
                    helpers.cond_print("Key not found in past query")
                copy_past[i] = extracted_attributes[i]
                copy_past2[i] = attribute_keyword_mapping[i]
            elif extracted_attributes[i]['meta']['followup_type'] == 'replace_to_remove':
                helpers.cond_print("Not going to do operation with to_remove")
                continue
            elif extracted_attributes[i]['meta']['followup_type'] == 'nothing':
                continue
            elif extracted_attributes[i]['meta']['followup_type'] == 'remove_only':
                if 'filter' not in task_map.keys():
                    copy_past_copy = dict()
                    copy_past2_copy = dict()
                    for j in copy_past.keys():
                        if j == i:
                            copy_past_copy[j] = extracted_attributes[i]
                            copy_past2_copy[j] = extracted_attributes[i]
                    copy_past = copy_past_copy
                    copy_past2 = copy_past2_copy
        return copy_past, copy_past2


    def get_followuptype_attribute_from_dependencies(self, dependencies):
        encodeable_attributes = self.nl4dv_instance.attribute_genie_instance.get_encodeable_attributes()
        # Initialize the output
        task_map = dict()
        operator = None
        _attributes = dict()
        keyword, operator_phrase = None, None
        if dependencies is not None:
            for dep_index, dep in enumerate(dependencies[0]):
                # OPTIONAL:
                if dep[1] in ['dobj', 'xcomp', 'amod']:
                    if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[2][0] in self.nl4dv_instance.followup_keyword_map:
                        operator_phrase = dep[2][0]
                        keyword = dep[0][0]

                    if dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[0][0] in self.nl4dv_instance.followup_keyword_map:
                        operator_phrase = dep[0][0]
                        keyword = dep[2][0]
                if keyword is not None and operator_phrase is not None:
                    followup_type = self.nl4dv_instance.followup_keyword_map[operator_phrase][0][0]
                    # These tasks requires at most 1 Attribute to make sense
                    if followup_type in ["addition", "removal"]:
                        operator = self.nl4dv_instance.followup_keyword_map[operator_phrase][0][1]
                        _keyword = self.nl4dv_instance.special_keyword_map_for_tasks[keyword]
                        _attributes_list = list(self.nl4dv_instance.keyword_attribute_mapping[_keyword].keys())
                        for attribute_key in _attributes_list:
                            if followup_type in ['removal']:
                                _attributes[attribute_key] = 'remove'
                            elif followup_type in ['addition']:
                                _attributes[attribute_key] = 'add'
                    keyword, operator_phrase = None, None
            k1, k2, operator_phrase = None, None, None
            for dep_index, dep in enumerate(dependencies[0]):
                if dep[1] in ['dobj', 'xcomp']:
                    if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[2][0] in self.nl4dv_instance.followup_keyword_map:
                        operator_phrase = dep[2][0]
                        k1 = dep[0][0]
                    if dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[0][0] in self.nl4dv_instance.followup_keyword_map:
                        operator_phrase = dep[0][0]
                        k1 = dep[2][0]

                if dep[1] in ['nmod', 'case']:
                    if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks and k1 == dep[0][0]:
                        k2 = dep[2][0]
                    if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[2][1] in ["IN"]:
                        k2 = dep[0][0]

                if k1 is not None and operator_phrase is not None and k2 is not None:
                    _k1 = self.nl4dv_instance.special_keyword_map_for_tasks[k1]
                    _k2 = self.nl4dv_instance.special_keyword_map_for_tasks[k2]
                    followup_type = self.nl4dv_instance.followup_keyword_map[operator_phrase][0][0]
                    # These tasks requires at most 1 Attribute to make sense
                    if followup_type in ["replacement"]:
                        relevant_attributes = []
                        operator = self.nl4dv_instance.followup_keyword_map[operator_phrase][0][1]
                        for k in [_k1, _k2]:
                            if k in self.nl4dv_instance.keyword_attribute_mapping:
                                relevant_attributes.extend(list(self.nl4dv_instance.keyword_attribute_mapping[k].keys()))
                                #DEBUG ELIF: WILL IT EVER BE CALLED?
                            elif self.nl4dv_instance.porter_stemmer_instance.stem(k) in self.nl4dv_instance.keyword_attribute_mapping:
                                relevant_attributes.extend(list(self.nl4dv_instance.keyword_attribute_mapping[self.nl4dv_instance.porter_stemmer_instance.stem(k)].keys()))
                        _attributes[list(self.nl4dv_instance.keyword_attribute_mapping[_k1].keys())[0]] = "replace_to_remove"
                        _attributes[list(self.nl4dv_instance.keyword_attribute_mapping[_k2].keys())[0]] = "replace_to_add"
                    k1, k2, operator_phrase = None, None, None
            if operator is None:
                operator, _attributes = self.get_implicit_followuptype_attribute_from_dependencies(dependencies)
        return [operator, _attributes]


    def get_followuptype_task_from_dependencies(self, dependencies, query_ngrams):
        encodeable_attributes = self.nl4dv_instance.attribute_genie_instance.get_encodeable_attributes()
        # Initialize the output
        task_map = dict()
        # Infer Tasks from the EXPLICIT
        keyword, operator_phrase = None, None
        for dep_index, dep in enumerate(dependencies[0]):
            # OPTIONAL:
            if dep[1] in ['amod', 'nsubj', 'nmod', 'dobj', 'compound', 'dep']:
                if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[2][0] in self.nl4dv_instance.task_keyword_map:
                    operator_phrase = dep[2][0]
                    keyword = dep[0][0]

                if dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[0][0] in self.nl4dv_instance.task_keyword_map:
                    operator_phrase = dep[0][0]
                    keyword = dep[2][0]

            if operator_phrase is not None and keyword is not None:
                task = self.nl4dv_instance.task_keyword_map[operator_phrase][0][0]
                if task not in task_map:
                    task_map[task] = list()
                # These tasks requires at most 1 Attribute to make sense
                if task in ["derived_value", "find_extremum", "sort"]:
                    operator = self.nl4dv_instance.task_keyword_map[operator_phrase][0][1]
                    final_followup_type, final_followup_token = None, None

                    if task == "sort":
                        for ngram in query_ngrams:
                            if ngram in self.nl4dv_instance.task_keyword_map:
                                if self.nl4dv_instance.task_keyword_map[ngram][0][0] == "sort":
                                    operator = self.nl4dv_instance.task_keyword_map[ngram][0][1]

                    for ngram in query_ngrams:
                        if ngram in self.nl4dv_instance.followup_keyword_map:
                            final_followup_type = self.nl4dv_instance.followup_keyword_map[ngram][0][1]

                    if final_followup_type is None:
                        if task not in self.nl4dv_instance.past_extracted_tasks:
                            final_followup_type = 'add'
                        else:
                            final_followup_type = 'nothing'

                    _keyword = self.nl4dv_instance.special_keyword_map_for_tasks[keyword]
                    _attributes = list(self.nl4dv_instance.keyword_attribute_mapping[_keyword].keys())
                    _tasks = self.nl4dv_instance.task_genie_instance.generate_tasks(task_name=task,
                                                 attributes=_attributes,
                                                 operator_phrase=operator_phrase,
                                                 query_phrase=[_keyword],
                                                 operator=operator,
                                                 values=[],
                                                 inference_type='explicit',
                                                 followup_type=final_followup_type,
                                                 allow_subset=False
                                                 )
                    for _task in _tasks:
                        if _task not in task_map[task]:
                            task_map[task].append(_task)
                keyword, operator_phrase = None, None

        keyword, amount, operator_phrase, has_negation, negation_phrase = None, None, None, False, None
        for dep_index, dep in enumerate(dependencies[0]):

            if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[2][1] == 'CD':
                keyword = dep[0][0]
                amount = dep[2][0]

            if dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[0][1] == 'CD':
                keyword = dep[2][0]
                amount = dep[0][0]

            if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[2][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                has_negation = True
                negation_phrase = dep[2][0]

            if dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[0][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                has_negation = True
                negation_phrase = dep[0][0]

            if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[2][0] in self.nl4dv_instance.task_keyword_map:
                operator_phrase = dep[2][0]
                keyword = dep[0][0]

            if dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[0][0] in self.nl4dv_instance.task_keyword_map:
                operator_phrase = dep[0][0]
                keyword = dep[2][0]

            if dep[0][1] == 'CD' and dep[2][0] in self.nl4dv_instance.task_keyword_map:
                operator_phrase = dep[2][0]
                amount = dep[0][0]

            if dep[2][1] == 'CD' and dep[0][0] in self.nl4dv_instance.task_keyword_map:
                operator_phrase = dep[0][0]
                amount = dep[2][0]

            if dep[0][1] == "CD" and dep[2][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                has_negation = True
                negation_phrase = dep[2][0]

            if dep[2][1] == "CD" and dep[0][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                has_negation = True
                negation_phrase = dep[0][0]

            if keyword is not None and amount is not None and operator_phrase is not None:
                if self.nl4dv_instance.task_keyword_map[operator_phrase][0][0] == "filter" and operator_phrase != "between":
                    operator = self.nl4dv_instance.task_keyword_map[operator_phrase][0][1]
                    if has_negation:
                        if operator == "GT":
                            operator = "LT"
                        elif operator == "LT":
                            operator = "GT"
                        operator_phrase = negation_phrase + " " + operator_phrase

                    task = 'filter'
                    if task not in task_map:
                        task_map[task] = list()

                    _keyword = self.nl4dv_instance.special_keyword_map_for_tasks[keyword]
                    _attributes = list(self.nl4dv_instance.keyword_attribute_mapping[_keyword].keys())
                    _value = self.nl4dv_instance.task_genie_instance.get_attributes_values(_attributes[0], amount)
                    _tasks = self.nl4dv_instance.task_genie_instance.generate_tasks(task_name=task,
                                                 attributes=_attributes,
                                                 operator_phrase=operator_phrase,
                                                 query_phrase=[_keyword],
                                                 operator=operator,
                                                 values=[_value],
                                                 inference_type='explicit',
                                                 followup_type = 'nothing',
                                                 allow_subset=False)

                    for _task in _tasks:
                        if _task not in task_map[task]:
                            task_map[task].append(_task)

                    keyword, amount, operator_phrase, has_negation, negation_phrase = None, None, None, False, None

        # Case 3: Numeric + Temporal FILTERS e.g '...between 5 and 10', '...between 2017/12/1 and 2018/1/1'
        keyword, from_amount, to_amount, operator_phrase, has_negation, negation_phrase = None, None, None, None, False, None
        for dep_index, dep in enumerate(dependencies[0]):
            if dep[0][1] == 'CD' and dep[1] == 'conj' and dep[2][1] == 'CD':
                from_amount = dep[0][0]
                to_amount = dep[2][0]

            if dep[0][1] == 'CD' and dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks:
                keyword = dep[2][0]

            if dep[2][1] == 'CD' and dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks:
                keyword = dep[0][0]


            if dep[0][1] == "CD" and dep[2][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                has_negation = True
                negation_phrase = dep[2][0]

            if dep[2][1] == "CD" and dep[0][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                has_negation = True
                negation_phrase = dep[0][0]

            if dep[0][1] == 'CD' and dep[2][0] in self.nl4dv_instance.task_keyword_map:
                operator_phrase = dep[2][0]

            if dep[2][1] == 'CD' and dep[0][0] in self.nl4dv_instance.task_keyword_map:
                operator_phrase = dep[0][0]

            if keyword is not None and from_amount is not None and to_amount is not None and operator_phrase is not None:
                if self.nl4dv_instance.task_keyword_map[operator_phrase][0][0] == "filter" and operator_phrase == "between":
                    operator = self.nl4dv_instance.task_keyword_map[operator_phrase][0][1]
                    if has_negation:
                        # NOT RANGE
                        operator = "NOT" + " " + operator
                        operator_phrase = negation_phrase + " " + operator_phrase

                    task = 'filter'
                    if task not in task_map:
                        task_map[task] = list()

                    _keyword = self.nl4dv_instance.special_keyword_map_for_tasks[keyword]
                    _attributes = list(self.nl4dv_instance.keyword_attribute_mapping[_keyword].keys())
                    _from_value = self.nl4dv_instance.task_genie_instance.get_attributes_values(_attributes[0], from_amount)
                    _to_value = self.nl4dv_instance.task_genie_instance.get_attributes_values(_attributes[0], to_amount)
                    _tasks = self.nl4dv_instance.task_genie_instance.generate_tasks(task_name=task,
                                                 attributes=_attributes,
                                                 operator_phrase=operator_phrase,
                                                 query_phrase=[_keyword],
                                                 operator=operator,
                                                 values=[_from_value, _to_value],
                                                 inference_type='explicit',
                                                 followup_type = 'nothing',
                                                 allow_subset=False)

                    for _task in _tasks:
                        if _task not in task_map[task]:
                            task_map[task].append(_task)

                    keyword, amount, operator_phrase, has_negation, negation_phrase = None, None, None, False, None
        if not task_map:
            final_followup_type = None
            _attributes = None
            operator_phrase = None
            operator = None
            ambiguity = []
            for ngram in query_ngrams:
                if ngram in self.nl4dv_instance.task_keyword_map:
                    if self.nl4dv_instance.task_keyword_map[ngram][0][0] in ["sort", "derived_value", "find_extremum"]:
                        ngram_flag = True
                        for key in self.nl4dv_instance.extracted_attributes.keys():
                            if ngram in key.lower():
                                ngram_flag = False
                        if ngram_flag:
                            operator = self.nl4dv_instance.task_keyword_map[ngram][0][1]
                            operator_phrase = ngram
                    if ngram in self.nl4dv_instance.followup_keyword_map:
                        final_followup_type = self.nl4dv_instance.followup_keyword_map[ngram][0][1]

                    if operator is not None and operator_phrase is not None:
                        task = self.nl4dv_instance.task_keyword_map[operator_phrase][0][0]
                        if task not in task_map:
                            task_map[task] = list()
                        if final_followup_type is None:
                            if task not in self.nl4dv_instance.past_extracted_tasks:
                                final_followup_type = 'add'
                            else:
                                final_followup_type = 'nothing'
                        q_attr = list()
                        for attr in self.nl4dv_instance.extracted_attributes.keys():
                            if self.nl4dv_instance.extracted_attributes[attr]['meta']['dataType'] == 'Q':
                                q_attr.append(attr)
                        if len(q_attr) == 1:
                            _attributes = q_attr
                        if _attributes is not None:
                            _tasks = self.nl4dv_instance.task_genie_instance.generate_tasks(task_name=task,
                                                                 attributes=_attributes,
                                                                 operator_phrase=operator_phrase,
                                                                 query_phrase=[],
                                                                 operator=operator,
                                                                 values=[],
                                                                 inference_type='explicit',
                                                                 followup_type=final_followup_type,
                                                                 allow_subset=False
                                                                 )
                        for _task in _tasks:
                                if _task not in task_map[task]:
                                    task_map[task].append(_task)
        return task_map

    def implicit_replacement(self, followup_dict, past_attributes):
        operator, attributes = followup_dict[0], followup_dict[1]
        key_to_remove = None
        if len(past_attributes.keys()) == 1:
            key_to_remove = list(past_attributes.keys())[0]
            attributes[key_to_remove] = "replace_to_remove"
        else:
            key_to_remove = random.choice(list(past_attributes))
            attributes[key_to_remove] = "replace_to_remove"
        return key_to_remove, [operator, attributes]


    def followup_conversation_task(self, task_map, past_task_map):
        copy_past = past_task_map.copy()
        new_copy_past = dict()
        for i in task_map.keys():
            if i in copy_past:
                past_task_list = copy_past[i]
                new_task_list = task_map[i]
                result = []
                if i == "sort":
                    for task in new_task_list:
                        if task["followup_type"] != "remove":
                            result = new_task_list
                elif i == "derived_value":
                    derived_value_list = list()
                    for task in new_task_list:
                        for past_task in past_task_list:
                            if past_task["operator"] != task["operator"] and set(task["attributes"]) == set(past_task["attributes"]) and task["followup_type"] != 'remove':
                                derived_value_list.append(dict(task))
                    copy_past[i] = derived_value_list
                elif i == "find_extremum":
                    find_extremum_list = list()
                    for task in new_task_list:
                        for past_task in past_task_list:
                            if past_task["operator"] != task["operator"] and set(task["attributes"]) == set(past_task["attributes"]) and task["followup_type"] != 'remove':
                                find_extremum_list.append(dict(task))
                    copy_past[i] = find_extremum_list
                else:
                    past_task_list.extend(new_task_list)
                    for task in past_task_list:
                        if task not in result:
                            result.append(task)
                    copy_past[i] = result
            else:
                copy_past[i] = task_map[i]
        for i in copy_past.keys():
            flag = True
            for task in copy_past[i]:
                if task['inferenceType'] == 'implicit':
                    flag = False

            if flag:
                new_copy_past[i] = copy_past[i]

        return new_copy_past

    def get_implicit_followuptype_attribute_from_dependencies(self, dependencies):
        encodeable_attributes = self.nl4dv_instance.attribute_genie_instance.get_encodeable_attributes()

        # Initialize the output
        task_map = dict()
        operator = None
        _attributes = dict()
        implicit_keyword = None

        if dependencies is not None:
            for dep_index, dep in enumerate(dependencies[0]):
                if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_followup.keys():
                    if operator is None or operator == "add":
                        operator = self.nl4dv_instance.special_keyword_map_for_followup[dep[0][0]]
                        implicit_keyword = dep[0][0]
                elif dep[2][0] in self.nl4dv_instance.special_keyword_map_for_followup.keys():
                    if operator is None or operator == "add":
                        operator = self.nl4dv_instance.special_keyword_map_for_followup[dep[2][0]]
                        implicit_keyword = dep[2][0]

            if operator == "remove":
                keyword = None
                for dep_index, dep in enumerate(dependencies[0]):
                    if dep[1] in ['advmod', 'xcomp', 'nmod', 'dobj']:
                        if dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks:
                            keyword = dep[2][0]
                        elif dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks:
                            keyword = dep[0][0]
                    if keyword is not None:
                        # These tasks requires at most 1 Attribute to make sense
                        _keyword = self.nl4dv_instance.special_keyword_map_for_tasks[keyword]
                        _attributes = list(self.nl4dv_instance.keyword_attribute_mapping[_keyword].keys())
                        davo = dict()
                        for attribute_key in _attributes:
                            if _keyword in attribute_key.lower() and operator in ['remove', 'remove_only'] and implicit_keyword == 'only':
                                davo[attribute_key] = 'remove_only'
                            else:
                                davo[attribute_key] = 'remove'
                        _attributes = davo
                keyword = None
            elif operator == "replace":
                k1, k2 = None, None
                for dep_index, dep in enumerate(dependencies[0]):
                    if dep[1] in ['compound', 'dobj']:
                        if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[2][0] in self.nl4dv_instance.special_keyword_map_for_followup:
                            k1 = dep[0][0]
                        if dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[0][0] in self.nl4dv_instance.special_keyword_map_for_followup:
                            k1 = dep[2][0]
                for dep_index, dep in enumerate(dependencies[0]):
                    if dep[1] in ['amod', 'acl', 'nsubj']:
                        if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks:
                            k2 = dep[0][0]
                        if dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks:
                            k2 = dep[2][0]
                for dep_index, dep in enumerate(dependencies[0]):
                    if dep[1] in ['amod', 'compound']:
                        if dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks and dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks:
                            k1 = dep[0][0]
                            k2 = dep[2][0]

                if k1 is not None and k2 is not None:
                    _k1 = self.nl4dv_instance.special_keyword_map_for_tasks[k1]
                    _k2 = self.nl4dv_instance.special_keyword_map_for_tasks[k2]

                    relevant_attributes = []
                    for k in [_k1, _k2]:
                        if k in self.nl4dv_instance.keyword_attribute_mapping:
                            relevant_attributes.extend(list(self.nl4dv_instance.keyword_attribute_mapping[k].keys()))
                        elif self.nl4dv_instance.porter_stemmer_instance.stem(k) in self.nl4dv_instance.keyword_attribute_mapping:
                            relevant_attributes.extend(list(self.nl4dv_instance.keyword_attribute_mapping[self.nl4dv_instance.porter_stemmer_instance.stem(k)].keys()))
                    davo = dict()
                    for attribute_key in relevant_attributes:
                        if _k1 in attribute_key.lower():
                            davo[attribute_key] = 'replace_to_remove'
                        elif _k2 in attribute_key.lower():
                            davo[attribute_key] = 'replace_to_add'
                    _attributes = davo

                    k1, k2 = None, None
                elif k2 is not None:
                    _k2 = self.nl4dv_instance.special_keyword_map_for_tasks[k2]
                    _attributes = list(self.nl4dv_instance.keyword_attribute_mapping[_k2].keys())
                    davo = dict()
                    for attribute_key in _attributes:
                        if attribute_key.lower() == _k2 and operator in ['replace']:
                            davo[attribute_key] = 'replace_to_add'
                    _attributes = davo
                    k2 = None
            elif operator == "add":
                keyword = None
                for dep_index, dep in enumerate(dependencies[0]):
                    if dep[1] in ['dobj', 'xcomp', 'amod', 'advmod']:
                            if dep[2][0] in self.nl4dv_instance.special_keyword_map_for_tasks:
                                keyword = dep[2][0]
                            elif dep[0][0] in self.nl4dv_instance.special_keyword_map_for_tasks:
                                keyword = dep[0][0]
                if keyword is not None:
                    # These tasks requires at most 1 Attribute to make sense
                    _keyword = self.nl4dv_instance.special_keyword_map_for_tasks[keyword]
                    _attributes = list(self.nl4dv_instance.keyword_attribute_mapping[_keyword].keys())
                    davo = dict()
                    for attribute_key in _attributes:
                        if _keyword in attribute_key.lower() and operator in ['add']:
                            davo[attribute_key] = 'add'
                    _attributes = davo
                keyword = None
        return [operator, _attributes]

    def followup_filter_task(self, task_map, past_task_map, query_ngrams):
        keyword = list()
        for ngram in query_ngrams:
            for filter_keyword in self.nl4dv_instance.filter_followup_keyword_map.keys():
                if query_ngrams[ngram]["lower"] == filter_keyword:
                    keyword.append(self.nl4dv_instance.filter_followup_keyword_map[filter_keyword][0])
        garbage = list()
        if 'filter' in task_map.keys():
            task_filter_copy = list()
            task_filter = task_map['filter']
            if 'filter' not in past_task_map.keys():
                if 'remove_addition' in keyword:
                    val_list = list()
                    for filter_curr in task_filter:
                        if filter_curr['operator'] == 'LT':
                            filter_curr['operator'] = 'GT'
                        elif filter_curr['operator'] == 'GT':
                            filter_curr['operator'] = 'LT'
                        elif filter_curr['operator'] == 'RANGE':
                            filter_curr['operator'] = 'NOT RANGE'

                        else:
                            filter_curr_val = filter_curr['values']
                            val_list += filter_curr_val
                            domain_values = list(self.nl4dv_instance.data_genie_instance.data_attribute_map[filter_curr['attributes'][0]]['domain'])
                            inclusion_list = list()
                            for val in domain_values:
                                if val not in val_list:
                                    inclusion_list.append(val)
                            filter_curr['values'] = inclusion_list
                if "add_ambiguous" in keyword:
                    task_map['filter'][0]['ambiguity'] = ['add_or_remove']
                past_task_map.update(task_map)
                return past_task_map
            else:
                past_task_filter = past_task_map['filter']
                task_filter_copy = list()
                if 'remove_addition' in keyword:
                    val_list = list()

                    for filter_curr in task_filter:
                        if filter_curr['operator'] == 'LT':
                            filter_curr['operator'] = 'GT'
                        elif filter_curr['operator'] == 'GT':
                            filter_curr['operator'] = 'LT'
                        elif filter_curr['operator'] == 'RANGE':
                            filter_curr['operator'] = 'NOT RANGE'
                        else:
                            filter_curr_val = filter_curr['values']
                            val_list += filter_curr_val
                            domain_values = list(self.nl4dv_instance.data_genie_instance.data_attribute_map[filter_curr['attributes'][0]]['domain'])
                            inclusion_list = list()
                            for val in domain_values:
                                if val not in val_list:
                                    inclusion_list.append(val)
                            filter_curr["values"] = inclusion_list
                if 'remove' in keyword:
                    del past_task_map['filter']
                    return past_task_map
                for filter_curr in task_filter:
                    attributes = filter_curr['attributes'][0]
                    index = None
                    for past_filter_ind in range(0, len(past_task_filter)):
                        if attributes == past_task_filter[past_filter_ind]['attributes'][0]:
                            index = past_filter_ind
                    if index is not None and "add" not in keyword and "remove_addition" not in keyword:
                        if filter_curr['values'] != past_task_filter[index]['values']:
                            replacement_vals = list()
                            for val in filter_curr['values']:
                                if val not in past_task_filter[index]['values']:
                                    replacement_vals.append(val)
                            filter_curr["values"] = replacement_vals
                            garbage.append(past_task_filter[index])
                            past_task_filter[index] = filter_curr
                            past_task_map['filter'] = past_task_filter
                    elif index is not None and "add" in keyword:
                        if "add_ambiguous" in keyword:
                            past_task_filter[index]['ambiguity'] = ['add_or_remove']
                        past_task_filter[index]['queryPhrase'] += filter_curr['queryPhrase']
                        past_task_filter[index]['values'] += filter_curr['values']
                        past_task_map['filter'] = past_task_filter
                    elif index is not None and "remove_addition" in keyword:
                        past_task_filter[index]['queryPhrase'] += filter_curr['queryPhrase']
                        intersection_list = list(set(past_task_filter[index]['values']).intersection(filter_curr['values']))
                        past_task_filter[index]['values'] = intersection_list
                    elif index is None:
                        past_task_filter.append(filter_curr)
                        past_task_map['filter'] = past_task_filter
                return past_task_map
        else:
            if 'remove' in keyword:
                del past_task_map['filter']
            return past_task_map

    def remove_old_tasks(self, task_map):
        for task_key in task_map:
            new_task_list = []
            for task in task_map[task_key]:
                preservation = True
                for attr in task['attributes']:
                    if attr not in self.nl4dv_instance.extracted_attributes:
                        preservation = False
                if preservation:
                    new_task_list.append(task)
            task_map[task_key] = new_task_list
        return task_map

    def merge_update(self, past_extracted_attributes, extracted_attributes, past_attribute_keyword_mapping, attribute_keyword_mapping):
        for i in past_extracted_attributes.keys():
            if i not in extracted_attributes.keys():
                extracted_attributes[i] = past_extracted_attributes[i]
            else:
                continue
        for i in past_attribute_keyword_mapping.keys():
            if i not in attribute_keyword_mapping.keys():
                attribute_keyword_mapping[i] = past_attribute_keyword_mapping[i]
            else:
                continue
        return extracted_attributes, attribute_keyword_mapping











    #create global list of objects and dictionary of lists, so dictionary of lists has elements, and the elements are the indices in the global list

    #tree vs adjacency list for NL4DV (what operations do we want to support, add, remove, get, modify)

    #pick one task (add attribute, and implicit task)












from nl4dv.utils import constants, helpers
import re
import nltk
from nltk import word_tokenize
from si_prefix import si_parse
from nl4dv.querygenie import QueryGenie
from nl4dv.attributegenie import AttributeGenie
from nl4dv.taskgenie import TaskGenie
from nl4dv.visgenie import VisGenie

class UpdateGenie:
    def __init__(self, nl4dv_instance):
        self.nl4dv_instance = nl4dv_instance
        self.dialog = False


    def update_vis(self, dialog_id, query_id):
        self.nl4dv_instance.ambiguity_copy = self.nl4dv_instance.ambiguities.copy()
        self.task_map = dict()
        for task_type in self.nl4dv_instance.extracted_tasks:
            self.task_map[task_type] = list()
            for task in self.nl4dv_instance.extracted_tasks[task_type]:
                if task['inferenceType'] == 'explicit':
                    self.task_map[task_type].append(task)

        if bool(self.nl4dv_instance.ambiguities['value']):
            for key in self.nl4dv_instance.ambiguities['value']:
                if self.nl4dv_instance.ambiguities['value'][key]['selected'] is not None:
                    if 'selected_attr' in self.nl4dv_instance.ambiguities['value'][key].keys():
                        self.nl4dv_instance.ambiguities['value'][key]['selected'] = self.nl4dv_instance.ambiguities['value'][key]['selected_attr']
                    self.update_value_ambiguity()
                    self.nl4dv_instance.ambiguities['value'][key]['selected_attr'] = self.nl4dv_instance.ambiguities['value'][key]['selected']
                    self.nl4dv_instance.ambiguities['value'][key]['selected'] = 'NL4DV_Resolved'


        if bool(self.nl4dv_instance.ambiguities['attribute']):
            for key in self.nl4dv_instance.ambiguities['attribute']:
                if self.nl4dv_instance.ambiguities['attribute'][key]['selected'] is not None:
                    if 'selected_attr' in self.nl4dv_instance.ambiguities['attribute'][key].keys():
                        self.nl4dv_instance.ambiguities['attribute'][key]['selected'] = self.nl4dv_instance.ambiguities['attribute'][key]['selected_attr']
                    self.update_attribute_ambiguity()
                    self.nl4dv_instance.ambiguities['attribute'][key]['selected_attr'] = self.nl4dv_instance.ambiguities['attribute'][key]['selected']
                    self.nl4dv_instance.ambiguities['attribute'][key]['selected'] = 'NL4DV_Resolved'

        # At this stage, which attributes are encodeable?
        encodeable_attributes = self.nl4dv_instance.attribute_genie_instance.get_encodeable_attributes()
        task_map = self.nl4dv_instance.task_genie_instance.extract_implicit_tasks_from_attributes(self.task_map, encodeable_attributes, self.nl4dv_instance.dialog)
        self.nl4dv_instance.extracted_tasks = self.nl4dv_instance.task_genie_instance.filter_empty_tasks(task_map)

        final_encodeable_attributes = self.nl4dv_instance.attribute_genie_instance.update_encodeable_attributes_based_on_tasks()



        self.nl4dv_instance.vis_list = self.nl4dv_instance.vis_genie_instance.get_vis_list(attribute_list=final_encodeable_attributes)

        output = {
            'status': 'SUCCESS' if len(self.nl4dv_instance.vis_list) > 0 else 'FAILURE',
            'debug': {'execution_durations': self.nl4dv_instance.execution_durations},
            'query_raw': self.nl4dv_instance.query_raw,
            'query': self.nl4dv_instance.query_processed,
            'dataset': self.nl4dv_instance.data_url if self.nl4dv_instance.data_url else self.nl4dv_instance.data_value,
            'alias': self.nl4dv_instance.alias_url if self.nl4dv_instance.alias_url else self.nl4dv_instance.alias_value,
            'visList': self.nl4dv_instance.vis_list,
            'attributeMap': self.nl4dv_instance.extracted_attributes,
            'taskMap': self.nl4dv_instance.extracted_tasks,
            'followUpQuery': self.nl4dv_instance.dialog,
            'contextObj': None,
            'attributeMapping': self.nl4dv_instance.attribute_keyword_mapping,
            'followUpConfidence' : None,
            'ambiguity' : self.nl4dv_instance.ambiguities,
            'ambiguityCopy' : self.nl4dv_instance.ambiguity_copy
        }
        self.nl4dv_instance.conversation_genie_instance.update_query(dialog_id, query_id, output)
        output['dialogId'] = dialog_id
        output['queryId'] = query_id

        return output




    def update_attribute_ambiguity(self):
        ambiguous_phrases = list(self.nl4dv_instance.ambiguities['attribute'].keys())
        for phrase in ambiguous_phrases:
            chosen_attr = self.nl4dv_instance.ambiguities['attribute'][phrase]['selected']
            new_extracted_attr = dict()
            for attr in self.nl4dv_instance.extracted_attributes:
                if self.nl4dv_instance.extracted_attributes[attr]['queryPhrase'][0] == phrase:
                    if chosen_attr == attr:
                        new_extracted_attr[chosen_attr] = self.nl4dv_instance.extracted_attributes[attr]
                else:
                    new_extracted_attr[attr] = self.nl4dv_instance.extracted_attributes[attr]
            self.nl4dv_instance.extracted_attributes = new_extracted_attr
            phrase_dict = self.nl4dv_instance.keyword_attribute_mapping[phrase]
            for attr in list(phrase_dict.keys()):
                if attr != chosen_attr:
                    del phrase_dict[attr]
            self.nl4dv_instance.keyword_attribute_mapping[phrase] = phrase_dict

            for attr in list(self.nl4dv_instance.attribute_keyword_mapping.keys()):
                if attr != chosen_attr:
                    query_phrase = list(self.nl4dv_instance.attribute_keyword_mapping[attr].keys())[0]
                    if query_phrase == phrase:
                        del self.nl4dv_instance.attribute_keyword_mapping[attr]


    def update_value_ambiguity(self):
        ambiguous_phrases = list(self.nl4dv_instance.ambiguities['value'].keys())
        chosen_values = list()
        for phrase in ambiguous_phrases:
            chosen_val = self.nl4dv_instance.ambiguities['value'][phrase]['selected']
            task_filter = self.task_map['filter']
            for index in range(0, len(task_filter)):
                if phrase in task_filter[index]['queryPhrase']:
                    chosen_values.append(chosen_val)
                    task_filter[index]['values'] = chosen_values
                else:
                    chosen_values = list()



    def auto_handle_attribute_ambiguity(self):
        self.nl4dv_instance.past_ambiguities_copy = self.nl4dv_instance.past_ambiguities.copy()
        attribute_ambiguity = self.nl4dv_instance.past_ambiguities['attribute']
        for query_phrase in attribute_ambiguity:
            if not attribute_ambiguity[query_phrase]['selected']:
                attr_list = attribute_ambiguity[query_phrase]['options']
                attr_keep = attr_list[0]
                attr_confidence = self.nl4dv_instance.past_extracted_attributes[attr_keep]['meta']['confidence']
                for attr in attr_list:
                    confidence = self.nl4dv_instance.past_extracted_attributes[attr]['meta']['confidence']
                    if confidence > attr_confidence:
                        attr_confidence = confidence
                        attr_keep = attr
                self.nl4dv_instance.past_ambiguities['attribute'][query_phrase]['selected'] = attr_keep

        ambiguous_phrases = list(self.nl4dv_instance.past_ambiguities['attribute'].keys())
        for phrase in ambiguous_phrases:
            if self.nl4dv_instance.past_ambiguities['attribute'][phrase]['selected']:
                chosen_attr = self.nl4dv_instance.past_ambiguities['attribute'][phrase]['selected']
                new_extracted_attr = dict()
                for attr in self.nl4dv_instance.past_extracted_attributes:
                    if self.nl4dv_instance.past_extracted_attributes[attr]['queryPhrase'][0] == phrase:
                        if chosen_attr == attr:
                            new_extracted_attr[chosen_attr] = self.nl4dv_instance.past_extracted_attributes[attr]
                    else:
                        new_extracted_attr[attr] = self.nl4dv_instance.past_extracted_attributes[attr]
                self.nl4dv_instance.past_extracted_attributes = new_extracted_attr

                for attr in list(self.nl4dv_instance.past_attribute_keyword_mapping.keys()):
                    if attr != chosen_attr:
                        query_phrase = list(self.nl4dv_instance.past_attribute_keyword_mapping[attr].keys())[0]
                        if query_phrase == phrase:
                            del self.nl4dv_instance.past_attribute_keyword_mapping[attr]
                self.nl4dv_instance.past_ambiguities['attribute'][phrase]['selected'] = 'NL4DV_Resolved'










class PromptGenie:

    def __init__(self, nl4dv_instance):
        # nl4dv instance
        self.nl4dv_instance = nl4dv_instance
        self.prompt = """
        Consider the below JSON array of objects describing low-level analytic tasks (fundamental operations that users perform when interacting with data visualizations) as a list of their "Name", "Description" and "Pro Forma Abstract" (a concise summary outlining the main elements for a given natural language query), with "Examples" (example natural language queries about different datasets), "Attribute Data Types and Visual Encodings" (the data type of the column titles in the provided dataset along with the preferred visual encodings in the recommended visualization), "Attributes and Visual Encoding Descriptions", and "Recommended Visualizations":
        [
        {
        "Name": "Correlation",
        "Description": "Given a set of data cases and two attributes, determine useful relationships between the values of those attributes.",
        "Pro Forma Abstract": "What is the correlation between attributes X and Y over a given set S of data cases?",
        "Examples": ["Is there a correlation between carbohydrates and fat?", "Do different genders have a preferred payment method?", "What is the relationship between budget and gross?"],
        "taskMap Encoding": "correlation",
        "Attribute Data Types and Visual Encodings":
            [{
                "X axis": "Quantitative",
                "Y axis": "Quantitative",
                "Other Encoding": <choose from : ["Nominal", "Ordinal", "Quantitative", "Temporal"]>,
                "Sort": null,
                "Filter": null,
            }],
        "Attributes and Visual Encodings Description": "The "X axis" key indicates that the first attribute is used for the horizontal or x-axis of the visualization. Similarly the "Y axis" key is used for the vertical or y-axis of the visualization. The attributes in the "Other Encoding" key are optional and are to be used in encodings in color, shape, size, and opacity. The "Sort" and "Filter" keys should not be used in this task.",
        "Recommended Visualization":["Scatterplot"]
        },
        {
        "Name": "Derived Value",
        "Description": "Given a set of data cases, compute an aggregate numeric representation of those data cases.",
        "Pro Forma Abstract": "What is the value of aggregation function F over a given set S of data cases?",
        "Examples": ["What is the average calorie content of Post cereals?", "What is the gross income of all stores combined?", "How many manufacturers of cars are there?"],
        "taskMap Encoding": "derived_value",
        "Attribute Data Types and Visual Encodings":
            [{
                "X axis": "Nominal" or "Ordinal",
                "Y axis": "Quantitative",
                "Other Encoding": <choose from : ["Nominal", "Ordinal", "Quantitative", "Temporal"]>,
                "Sort": null,
                "Filter": null,
            }],
        "Attributes and Visual Encodings Description": "The "X axis" indicates that the first attribute is used for the horizontal or x-axis of the visualization. Similarly the "Y axis" is used for the vertical or y-axis of the visualization. The attributes in the "Other Encoding" key are optional and are to be used in encodings in color, shape, size, and opacity. The "Sort" and "Filter" keys should not be used in this task.",
        "Recommended Visualization" : ["Bar Chart"]
        },
        {
        "Name": "Filter",
        "Description": "Given some concrete conditions on attribute values, find data cases satisfying those conditions.",
        "Pro Forma Abstract": "Which data cases satisfy conditions {A, B, C, ...}?",
        "Examples": ["What Kellogg's cereals have high fiber?", "What comedies have won awards?", "Which funds underperformed the SP-500?",  "Show the visualization only for films that have grossed more than 100 million dollars"],
        "taskMap Encoding": "filter",
        "Attribute Data Types and Visual Encodings":
            [{
                "X axis": null
                "Y axis": null
                "Other Encoding": null,
                "Sort": null,
                "Filter": True,
            }],
        "Attributes and Visual Encodings Description":  "The "X axis", "Y axis", "Other Encoding", and "Sort" keys should not be used for this task. However, the "Filter" key should be used. As the "Filter" key is set to True, this indicates that the ensuing visualization must satisfy the conditions requested by the input natural language query. As such, the ensuing visualization must only display data points that satisfy these conditions. The attribute requested for the filter task can be any datatype (Quantiative, Nominal, Ordinal, or Temporal). If there is a "Filter" task detected in the input natural language query, please add the "transform" property in Vega-Lite to the Vega-Lite specification. This action will apply the filter specified in the natural language query. (Link to Vega-Lite "transform" property: https://vega.github.io/vega-lite/docs/filter.html)",
        "Visualization Recommendation":["Line Chart", "Scatter Plot", "Strip Plot", "Histogram", "Bar Chart", "Heatmap"]
        },
        {
        "Name": "Trend",
        "Description": "Trend is the direction of the data over time, which may be increasing, decreasing, or flat",
        "Pro Forma Abstract": "What is the direction of values for attribute(a) in the span of Time(t)?",
        "Examples": ["Is there a trend of increasing film length over the years?", "How have production budgets changed over the years?"],
        "taskMap Encoding": "trend",
        "Attribute Data Types and Visual Encodings":
            [{
                "X axis": "Temporal",
                "Y axis": "Quantitative",
                "Other Encoding": <choose from : ["Quantitative", "Nominal", "Ordinal"]>,
                "Sort": null,
                "Filter": null,
            }],
        "Attributes and Visual Encodings Description": "The "X axis" indicates that the first attribute is used for the horizontal or x-axis of the visualization. Similarly the "Y axis" is used for the vertical or y-axis of the visualization. Attributes in the curly brackets {} are optional and are to be used in the following encodings: color, shape, size, and opacity. The "Sort" and "Filter" keys are not to be used for this task.",
        "Visualization Recommendation":["Line Chart"]
        },
        {
        "Name": "Distribution",
        "Description": "Given a set of data cases and a quantitative attribute of interest, characterize the distribution of that attribute’s values over the set",
        "Pro Forma Abstract": "What is the distribution of values of attribute A in a set S of data cases?",
        "Examples": ["What is the age distribution of shoppers?", "What is the distribution of carbohydrates in cereals?", "What is the count of movies for each genre?"],
        "taskMap Encoding": "distribution",
        "Attribute Data Types and Visual Encodings":
            [{
                "X axis": "Quantitative" or "Nominal" or "Ordinal",
                "Y axis": "Quantitative",
                "Other Encoding": <choose from : ["Nominal", "Ordinal", "Quantitative", "Temporal"]>,
                "Sort": null,
                "Filter": null,
            }],
        "Attributes Description": "The "X axis" indicates that the first attribute is used for the horizontal or x-axis of the visualization. Similarly the "Y axis" is used for the vertical or y-axis of the visualization. Attributes in the curly brackets {} are optional and are to be used in encodings in color, shape, size, and opacity.",
        "Visualization Recommendation":["Histogram"]
        },
        {
        "Name": "Sort",
        "Description": "Given a set of data cases, rank them according to some ordinal metric.",
        "Pro Forma Abstract": "What is the sorted order of a set S of data cases according to their value of attribute A?",
        "Examples": ["Order the cars by weight", "Rank the cereals by calories."],
        "taskMap Encoding": "sort",
        "Attribute Data Types and Visual Encodings": 
            [{
                "X axis": null
                "Y axis": null
                "Other Encoding": null,
                "Sort": True
                "Filter": null
            }],
        "Attributes and Visual Encodings Description": "The "X axis", "Y axis", "Other Encoding", and "Filter" keys should not be used for this task. However, the "Sort" key should be used. As the "Sort" key is set to True, this indicates that the ensuing visualization must satisfy the order/ranking requested by the input natural language query. The attribute requested for the sort task must be of the Quantitative datatype.",
        "Visualization Recommendation":["Bar Chart"]
        },
        {
        "Name": "Find Extremum",
        "Description": "Find data cases possessing an extreme value of an attribute over its range within the data set.",
        "Pro Forma Abstract": "What are the top/bottom N data cases with respect to attribute A?",
        "Examples": ["What is the car with the highest MPG?", "What director/film has won the most awards?","What Robin Williams film has the most recent release date?"],
        "taskMap Encoding": "find_extremum",
        "Attribute Data Types and Visual Encodings":
            [{
                "X axis": null
                "Y axis": null
                "Other Encoding": null,
                "Sort": True
                "Filter": null
            }],
        "Attributes and Visual Encodings Description": "The "X axis", "Y axis", "Other Encoding", and "Filter" keys should not be used for this task. However, the "Sort" key should be used. As the "Sort" key is set to True, this indicates that the ensuing visualization must satisfy the order/ranking requested by the input natural language query. The attribute requested for the sort task must be of the Quantitative datatype.",
        "Visualization Recommendation":["Bar Chart"]
        }
        ]`

        Also consider the below JSON array of objects describing low-level operations for follow-up queries as a list of their "Name", "Instruction", and "Examples":
        '[
        {
        "Name": "Add Attribute",
        "Instruction": "Given a previous analytic specification, add the attribute that is detected in the input natural language query to the previous specification.",
        "Examples": ["Add genre", "Group by genre", "Put genre as well"]
        },
        {
        "Name": "Remove Attribute",
        "Instruction": "Given a previous analytic specification, remove the attribute that is detected in the input natural language query to the previous specification.",
        "Examples": ["Remove genre", "Show only production budget"]
        },
        {
        "Name": "Replace Attribute",
        "Instruction": "Given a previous analytic specification, replace the attribute that is detected in the input natural language query and is also in the previous analytic specification with the attribute that is detected in the natural language query, but is not in the previous specification.",
        "Examples": ["Replace gross with imdb rating", "Show gross instead of imdb rating"]
        },
        {
        "Name": "Add Task",
        "Instruction": "Given a previous analytic specification, add the task that is detected in the input natural language query to the previous specification. The visualization specification should reflect this new task that is added from the natural language query. Also please keep the previous tasks from the previous specification in the new modified analytic specification.",
        "Examples": ["Sort by worldwide gross in the descending order", "Show only horror movies", "What are their average production budgets?"]
        },
        {
        "Name": "Remove Task",
        "Instruction": "Given a previous analytic specification, remove the task that is detected in the input natural language query to the previous specification.",
        "Examples": ["Undo the sorting", "Undo the average"]
        },
        {
        "Name": "Replace Task",
        "Instruction": "Given a previous analytic specification, replace the attribute that is detected in the input natural language query and is also in the previous analytic specification with the attribute that is detected in the natural language query, but is not in the previous specification. Also please keep the other tasks from the previous specification that will not be replaced in the new modified analytic specification.",
        "Examples": ["Replace average with sum", "Show me total instead of average", "What about the lowest instead of highest?"]
        },
        {
        "Name": "Replace Vis",
        "Instruction": "Given a previous analytic specification, replace the visualization type that is detected in the input natural language query and is also in the previous analytic specification with the visualization type that is detected in the natural language query, but is not in the previous specification.",
        "Examples": ["Replace the vis with a bar chart", "As a line chart now", "Show a tick plot instead"]
        }
        ]’

        If a field called **PREVIOUS ANALYTIC SPECIFICATION** appears below the natural language query, classify the below natural language query into the respective follow-up operations they map to. Utilize the previous analytic specification (including the attributeMap, taskMap, and visList) and modify this specification to reflect the changes specified and requested in the natural language query. Return the visualization type in the form of a Vega-Lite specification where it reads data from url: https://raw.githubusercontent.com/nl4dv/nl4dv/master/examples/assets/data/movies-w-year.csv.
        However, if there is no field called PREVIOUS ANALYTIC SPECIFICATION that is below the natural language query,  then using the above definitions, classify the below natural language queries into the respective analytic tasks they map to. There can be one or more analytic tasks detected in the input natural language query. Return the visualization type in the form of a Vega-Lite specification where it reads data from the url above. PLEASE ensure that the schema used in your Vega-Lite specification is https://vega.github.io/schema/vega-lite/v4.json.

        Here's a subset of the original dataset with actual columns and rows for reference.

        <INSERT DATASET HERE>

        If there is no PREVIOUS ANLAYTIC SPECIFICATION that appears below the query, detect any attributes, tasks, and visualizations in the dataset that the provided query references, and place the detected dataset columns in the attributeMap, taskMap and visList property of the JSON below. Each Query can have more than one task and visualization type they can map to. Each property in the "attributeMap" JSON should be populated with the extracted dataset column (e.g. "Worldwide Gross"). There can be multiple attributes, tasks, and visualizations that are detected, but make sure that each attribute, task, and visualization in the attributeMap, taskMap, and visList is unique. Put each attribute into the attributeMap, task into taskMap, and Visualization into visList JSON. There can also be multiple possible visualization specifications as well. For each possible visualization specification, you can include a dictionary that has the required contents to the "visList" list.
        Furthermore, note that the input natural language (NL) query can also use ambiguous language with partial references to data attributes. In these cases, the attributeMap also includes an "isAmbiguous" field. The field can either be True or False. Set the field to True if the queryPhrase could refer to other attributes in the dataset. Otherwise set the field to False. If there are ambiguous attributes detected, generate Vega-Lite specifications for each ambiguous attribute that encode the ambiguous attribute in the visualization. Furthermore, if there are no tasks detected in the NL query, infer the task that is best suited with the detected attributes' datatypes. Generate a visualization specification using this inferred task and detected attributes.
        Here is the JSON object that the response should be returned as if and only if there is no PREVIOUS ANALYTIC SPECIFICATION attached to the natural language query:

        {
        "query": <Add NL query that is being parsed here>,
        "query_raw": <Add NL query that is being parsed here>,
        "dataset": <Add dataset URL here>,
        "visList": [
        {
        "attributes": [List of dataset attributes detected],
        "queryPhrase": [<Keywords found in query that were used to detect the taskMap and the recommended visualization>],
        "visType": "None",
        "tasks": [Add the list of tasks detected here. Utilize the value from the "taskMap Encoding" key in the analytic task JSON array to populate this list],
        "inferenceType": <Can be one of two values: "explicit" or "implicit". Set the value to "explicit" if the visualization's "queryPhrase" explicitly references a visualization type. Otherwise set the value to "implicit".>
        "vlSpec": <Add the Vega-Lite specification of the visualization recommended here.>
        }
        ],
        "attributeMap": {
        <Dataset column that was detected (should have same value as "name")>: {
        "name": <Dataset column that was detected>,
        "queryPhrase": [<Keywords found in query that were used to detect the dataset attribute>]
        "encode": <Boolean value depending on if the attribute appears on either of the axes or color in the Vega-Lite specification. The boolean value should be output as true or false in all lowercase letters.>
        },
        "metric": <[Can be one of two values: "attribute_exact_match" or "attribute_similarity_match".  Set the value to "attribute_exact_match" if the attribute was found directly in the query. Set the value to "attribute_similarity_match" if the query uses a synonym for the attribute.]>,
        "inferenceType": <Can be one of two values: "explicit" or "implicit". Set the value to "explicit" if the attribute’s "queryPhrase" references an attribute name. Set the value to "implicit" if the queryPhrase directly references values found in the attribute’s values.>
        "isAmbiguous": <Can be either True or False. Set the field to True if the queryPhrase could refer other attributes in the dataset. Otherwise set the field to False.>
        "ambiguity": [<Populate this list with all the different attributes in the dataset that the queryPhrase can refer to  if isAmbiguous is set to True. Otherwise keep this list empty.]
        },
        "taskMap": {
        <Task that was detected. Utilize the value from the "taskMap Encoding" key in the analytic task JSON array to populate this key>: [
        {
        "task": <Task that was detected. You must utilize the value from the "taskMap Encoding" key in the analytic task JSON array to populate this key>,
        "queryPhrase": [<Keywords found in query that were used to detect the task>],
        "values": [<If the "Filter" task was detected, put the filter value in here>],
        "attributes": [<Populate with the attributes that the task is mapped to]>],
        "operator": "<Can be one of "IN", "GT", "EQ", "AVG", "SUM", "MAX", or "MIN". "GT" is greater than. "EQ" is equals. "GT" and "EQ" are used for quantitative filters. "IN" is used for nominal filters. "AVG" and "SUM" are used for derived value tasks. "SUM" is for summation and "AVG" is for average. "MAX" and "MIN" are to be used for the sort and find extremum tasks. "MAX" indicates that the highest value must be displayed first. "MIN" indicates that the lowest value must be displayed first. Keep the string empty otherwise.>"
        "inferenceType": <Can be one of two values: "explicit" or "implicit". Set the value to "explicit" if the "queryPhrase" directly requests for a task to be applied. Set the value to "implicit" if the task is derived implicitly from the "queryPhrase".>
        }
        ]
        }
        "visType": <Put the visualization type that was explicitly specified in the query. Put the string "None" if vis type was not specified in the query.>,
        "visQueryPhrase": <Keywords found in query that were used to detect the visType. Put the string "None" if vis type was not specified in the query.>
        }

        If there is a PREVIOUS ANALYTIC SPECIFICATION provided, then utilize it (including the attributeMap, taskMap, and visList) and modify this specification to reflect the changes that are specified and requested in the natural language query. Then return this modified JSON object as output. The Vega-Lite specification should also change and a new visualization should be generated as a result of the modification. Also, in the modified JSON object, include a new property called "followupType". This "followupType" should include the classified follow-up operation (add attribute, replace task, etc.) that was detected in the natural language query.
        Please do not add any extra prose to your response. I only want to see the JSON output.

        <INSERT QUERY HERE>
        """


from nl4dv.utils import constants, helpers
import re
import nltk
from nltk import word_tokenize
from si_prefix import si_parse
from nl4dv.querygenie import QueryGenie
from nl4dv.attributegenie import AttributeGenie
from nl4dv.taskgenie import TaskGenie
from nl4dv.visgenie import VisGenie

class AutoGenie:
    def __init__(self, nl4dv_instance):
        self.nl4dv_instance = nl4dv_instance
        self.dialog = False

    def analyze_followup_status(self, query, reserve_words, ignore_words):
        self.query_raw = query
        self.dialog = False

        self.query_processed = self.nl4dv_instance.query_genie_instance.process_query(self.query_raw)
        self.special_keyword_map_for_followup = dict()
        for k,v in self.nl4dv_instance.implicit_followup_keyword_map.items():
            if k in self.query_processed:
                self.query_processed = self.query_processed.replace(k, v[0][0])
                self.special_keyword_map_for_followup[v[0][0]] = [v[0][1], v[0][2]]
        reserve_words = self.nl4dv_instance.reserve_words
        ignore_words = self.nl4dv_instance.ignore_words


        self.query_tokens = self.nl4dv_instance.query_genie_instance.clean_query_and_get_query_tokens(self.query_processed, reserve_words, ignore_words, dialog=True)
        query_ngrams = self.nl4dv_instance.query_genie_instance.get_query_ngrams(' '.join(self.query_tokens))
        ngram_set = set()
        for ngram in query_ngrams:
            ngram_set.add(ngram)
            if query_ngrams[ngram]["lower"] in self.nl4dv_instance.followup_keyword_map:
                self.dialog = True
                return self.dialog, "high"
            elif query_ngrams[ngram]["lower"] in self.special_keyword_map_for_followup:
                ambiguity = self.special_keyword_map_for_followup[query_ngrams[ngram]["lower"]][1]
                if ambiguity == 'not ambiguous':
                    self.dialog = True
                    return self.dialog, "high"
                else:
                    for ngram_sub in query_ngrams:
                        if ngram_sub not in ngram_set:
                            if query_ngrams[ngram_sub]["lower"] in self.nl4dv_instance.followup_keyword_map:
                                self.dialog = True
                                return self.dialog, "high"
                            elif query_ngrams[ngram_sub]["lower"] in self.special_keyword_map_for_followup:
                                ambiguity = self.special_keyword_map_for_followup[query_ngrams[ngram_sub]["lower"]][1]
                                if ambiguity == 'not ambiguous':
                                    self.dialog = True
                                    return self.dialog, "high"
                    followup_type = self.special_keyword_map_for_followup[query_ngrams[ngram]["lower"]][0]
                    self.nl4dv_instance.extracted_attributes = self.nl4dv_instance.attribute_genie_instance.extract_attributes(query_ngrams)
                    self.nl4dv_instance.query_for_task_inference = self.nl4dv_instance.task_genie_instance.prepare_query_for_task_inference(self.query_processed)
                    self.nl4dv_instance.extracted_vis_type, self.nl4dv_instance.extracted_vis_token = self.nl4dv_instance.vis_genie_instance.extract_vis_type(query_ngrams)

                    self.nl4dv_instance.dependencies = self.nl4dv_instance.task_genie_instance.create_dependency_tree(self.nl4dv_instance.query_for_task_inference)
                    task_map = self.nl4dv_instance.task_genie_instance.extract_explicit_tasks_from_dependencies(self.nl4dv_instance.dependencies)
                    # Filters from Domain Values
                    task_map = self.nl4dv_instance.task_genie_instance.extract_explicit_tasks_from_domain_value(task_map)

                    # From the generated TaskMap, ensure that the task "keys" are NOT EMPTY LISTS
                    self.nl4dv_instance.extracted_tasks = self.nl4dv_instance.task_genie_instance.filter_empty_tasks(task_map)
                    if (bool(self.nl4dv_instance.extracted_tasks) + bool(self.nl4dv_instance.extracted_attributes) + bool(self.nl4dv_instance.extracted_vis_type)) == 1:
                        if self.nl4dv_instance.extracted_vis_type is not None:
                            recent_dialog_id = self.nl4dv_instance.conversation_genie_instance.most_recent_dialog_id[-1]

                            recent_query_id = self.nl4dv_instance.conversation_genie_instance.most_recent_query_id[-1]
                            context_obj = self.nl4dv_instance.conversation_genie_instance.all_dialogs[recent_dialog_id][int(recent_query_id)]
                            past_vis_type = context_obj['visList'][0]['visType']
                            if past_vis_type != self.nl4dv_instance.extracted_vis_type:
                                self.dialog = True
                                return self.dialog, "high"
                            else:
                                self.dialog = False
                                return self.dialog, "none"
                        elif self.nl4dv_instance.extracted_attributes is not None:
                            if len(self.nl4dv_instance.extracted_attributes.keys()) == 1:
                                self.dialog = True
                                return self.dialog, "low"
                            elif len(self.nl4dv_instance.extracted_attributes.keys()) == 2 and followup_type == "replace":
                                self.dialog = True
                                return self.dialog, "low"
                        elif self.nl4dv_instance.extracted_tasks is not None:
                            if len(self.nl4dv_instance.extracted_tasks.keys()) == 1:
                                self.dialog = True
                                return self.dialog, "low"


                        attr = list(self.nl4dv_instance.extracted_attributes.keys())
                        task_attr = list()
                        for task in self.nl4dv_instance.extracted_tasks.keys():
                            for task_curr in self.nl4dv_instance.extracted_tasks[task]:
                                attrs = task_curr['attributes']
                                task_attr.extend(attrs)

                        if set(attr) == set(task_attr):
                            self.dialog = True
                            return self.dialog, "low"
                    elif (bool(self.nl4dv_instance.extracted_tasks) + bool(self.nl4dv_instance.extracted_attributes) + bool(self.nl4dv_instance.extracted_vis_type)) == 0:
                        self.dialog = False
                        return self.dialog, "none"

        self.nl4dv_instance.extracted_attributes = self.nl4dv_instance.attribute_genie_instance.extract_attributes(query_ngrams)

        self.nl4dv_instance.query_for_task_inference = self.nl4dv_instance.task_genie_instance.prepare_query_for_task_inference(self.query_processed)
        self.nl4dv_instance.extracted_vis_type, self.nl4dv_instance.extracted_vis_token = self.nl4dv_instance.vis_genie_instance.extract_vis_type(query_ngrams)

        self.nl4dv_instance.dependencies = self.nl4dv_instance.task_genie_instance.create_dependency_tree(self.nl4dv_instance.query_for_task_inference)
        task_map = self.nl4dv_instance.task_genie_instance.extract_explicit_tasks_from_dependencies(self.nl4dv_instance.dependencies)
        # Filters from Domain Values
        task_map = self.nl4dv_instance.task_genie_instance.extract_explicit_tasks_from_domain_value(task_map)

        # From the generated TaskMap, ensure that the task "keys" are NOT EMPTY LISTS
        self.nl4dv_instance.extracted_tasks = self.nl4dv_instance.task_genie_instance.filter_empty_tasks(task_map)
        if (bool(self.nl4dv_instance.extracted_tasks) + bool(self.nl4dv_instance.extracted_attributes) + bool(self.nl4dv_instance.extracted_vis_type)) == 1:
            if self.nl4dv_instance.extracted_vis_type is not None:
                recent_dialog_id = self.nl4dv_instance.conversation_genie_instance.most_recent_dialog_id[-1]
                recent_query_id = self.nl4dv_instance.conversation_genie_instance.most_recent_query_id[-1]
                context_obj = self.nl4dv_instance.conversation_genie_instance.all_dialogs[recent_dialog_id][int(recent_query_id)]
                past_vis_type = context_obj['visList'][0]['visType']
                if past_vis_type != self.nl4dv_instance.extracted_vis_type:
                    self.dialog = True
                    return self.dialog, "high"
                else:
                    self.dialog = False
                    return self.dialog, "none"
            elif self.nl4dv_instance.extracted_tasks is not None:
                if len(self.nl4dv_instance.extracted_tasks.keys()) == 1:
                    if task_map_key in ['filter', 'derived_value', 'find_extremum', 'sort']:
                        self.dialog = True
                        return self.dialog, "low"
            elif self.nl4dv_instance.extracted_attributes is not None:

                if len(self.nl4dv_instance.extracted_attributes.keys()) == 1:
                    self.dialog = True
                    return self.dialog, "low"

        elif (bool(self.nl4dv_instance.extracted_tasks) + bool(self.nl4dv_instance.extracted_attributes) + bool(self.nl4dv_instance.extracted_vis_type)) == 0:
            self.dialog = False
            return self.dialog, "none"

        else:
            attr = list(self.nl4dv_instance.extracted_attributes.keys())

            task_attr = list()
            for task in self.nl4dv_instance.extracted_tasks.keys():
                if task in ['filter', 'derived_value', 'find_extremum', 'sort']:
                    for task_curr in self.nl4dv_instance.extracted_tasks[task]:
                        attrs = task_curr['attributes']
                        task_attr.extend(attrs)
            if set(attr) == set(task_attr):
                self.dialog = True
                return self.dialog, "low"


        return self.dialog, "none"





from setuptools import setup, find_packages
from os import path

# read the contents of the README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='nl4dv',
    version='3.0.0',
    author='Arpit Narechania',
    author_email='arpitnarechania@gmail.com',
    packages=find_packages(),
    include_package_data=True,
    setup_requires=['pytest-runner'],
    test_requires=['pytest', 'pytest-cov'],
    scripts=[],
    url='https://github.com/nl4dv/nl4dv',
    license='LICENSE',
    description='NL4DV is a Python toolkit that takes a natural language (NL) query about a given dataset as input and outputs a structured JSON object containing Data attributes, Analytic tasks, and Visualizations (Vega-Lite specifications).',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        "fuzzywuzzy~=0.8.1",
        "nltk~=3.5",
        "pytest~=3.10.1",
        "pytest-cov~=2.6.0",
        "python-Levenshtein~=0.25.1",
        "si-prefix~=1.3.3",
        "spacy~=3.0",
        "vega~=3.1",
        "inflect~=5.5.2",
        "openai~=0.28.1"
    ]
)



 # In-built Libraries
import os
import time
from collections import OrderedDict
import json

# Third-Party Libraries
import spacy
import openai
import pandas as pd
import re
from nltk.stem.porter import PorterStemmer
from nltk.parse.stanford import StanfordDependencyParser
from nltk.parse.corenlp import CoreNLPDependencyParser

from vega import VegaLite

# NL4DV Imports
from nl4dv.datagenie import DataGenie
from nl4dv.querygenie import QueryGenie
from nl4dv.attributegenie import AttributeGenie
from nl4dv.taskgenie import TaskGenie
from nl4dv.visgenie import VisGenie
from nl4dv.updategenie import UpdateGenie
from nl4dv.utils import helpers, constants, error_codes
from nl4dv.conversationgenie import ConversationGenie
from nl4dv.autogenie import AutoGenie
from nl4dv.updategenie import UpdateGenie
from nl4dv.promptgenie import PromptGenie

class NL4DV:
    """
    Class exposed to users to interact with the package. Exposes modules in the package via
    public methods

    """

    def __init__(self, data_url=None,
                 data_value=None,
                 alias_url=None,
                 alias_value=None,
                 label_attribute=None,
                 ignore_words=list(),
                 reserve_words=list(),
                 explicit_followup_keywords = dict(),
                 implicit_followup_keywords = dict(),
                 dependency_parser_config=None,
                 thresholds=None,
                 importance_scores=None,
                 attribute_datatype=None,
                 verbose=False,
                 debug=False,
                 processing_mode = 'semantic-parsing',
                 gpt_api_key = ''):

        # inputs
        self.data_url = data_url
        self.data_value = data_value
        self.alias_url = alias_url
        self.alias_value = alias_value
        self.label_attribute = label_attribute
        self.ignore_words = ignore_words
        self.reserve_words = reserve_words
        self.dependency_parser_config = dependency_parser_config
        self.verbose = verbose
        self.debug = debug
        self.processing_mode = processing_mode

        # Load constants: thresholds, mappings, scores
        self.vis_keyword_map = constants.vis_keyword_map
        self.task_keyword_map = constants.task_keyword_map
        if bool(explicit_followup_keywords):
            self.followup_keyword_map = explicit_followup_keywords
        else:
            self.followup_keyword_map = constants.followup_keyword_map
        if bool(implicit_followup_keywords):
            self.implicit_followup_keyword_map = implicit_followup_keywords
        else:
            self.implicit_followup_keyword_map = constants.implicit_followup_keyword_map
        self.filter_followup_keyword_map = constants.filter_followup_keyword_map
        self.followup_reserve_words = constants.followup_reserve_words
        self.match_scores = constants.match_scores
        self.match_thresholds = constants.match_thresholds

        # Initialize intermediate/output variables
        self.execution_durations = dict()
        self.query_raw = None
        self.query_processed = ""
        self.query_for_task_inference = ""
        self.query_tokens = list()
        self.query_ngrams = dict()
        self.extracted_vis_type = None
        self.extracted_vis_token = None
        self.extracted_tasks = OrderedDict()
        self.extracted_attributes = OrderedDict()
        self.vis_list = None
        self.dependencies = list()
        self.dependency_parser = None
        self.dependency_parser_instance = None

        # Others
        self.dialog = False
        openai.api_key = gpt_api_key

        # initialize porter stemmer instance
        self.porter_stemmer_instance = PorterStemmer()

        # Initialize internal Class Instances
        self.data_genie_instance = DataGenie(self)  # initialize a DataGenie instance.
        self.query_genie_instance = QueryGenie(self)  # initialize a QueryGenie instance.
        self.attribute_genie_instance = AttributeGenie(self)   # initialize a AttributeGenie instance.
        self.task_genie_instance = TaskGenie(self)  # initialize a TaskGenie instance.
        self.vis_genie_instance = VisGenie(self)   # initialize a VisGenie instance.
        self.conversation_genie_instance = ConversationGenie(self)
        self.autogenie_instance = AutoGenie(self)
        self.updategenie_instance = UpdateGenie(self) #initalize a FollowUpGenie instance
        self.promptgenie_instance = PromptGenie(self)



        # Set the dependency parser if config is not None
        if self.dependency_parser_config is not None:
            self.set_dependency_parser(dependency_parser_config)

        # Set the thresholds, e.g., string matching
        if thresholds is not None:
            self.set_thresholds(thresholds=thresholds)

        # Set the importance scores, i.e., weights assigned to different ways in which attributes, tasks, and vis are detected/inferred.
        if importance_scores is not None:
            self.set_importance_scores(scores=importance_scores)

        # Override the attribute datatypes
        if attribute_datatype is not None:
            self.set_attribute_datatype(attr_type_obj=attribute_datatype)

    # returns a VegaLite object of the best (1st) visualization after analyzing the query.
    def render_vis(self, query, dialog=None, dialog_id = None, query_id = None):
        # type: (str, str, str, str) -> VegaLite
        response = self.analyze_query(query=query, dialog=dialog, dialog_id=dialog_id, query_id=query_id)
        if len(response['visList']) == 0:
            print("No best Viz; please try again.")
            return VegaLite({})
        return VegaLite(response['visList'][0]['vlSpec'])

    # ToDo:- Discuss support for non-ascii characters? Fallback from unicode to ascii good enough?
    # ToDo:- Discuss ERROR Handling
    # ToDo:- Utilities to perform unit conversion (eg. seconds > minutes). Problem: Tedious to infer base unit from data. - LATER
    def analyze_query(self, query=None, dialog=None, debug=None, verbose=None, dialog_id = None, query_id = None):
        # type: (str, bool, bool, bool, str, str) -> dict
        if self.processing_mode == 'semantic-parsing':
            if not bool(self.conversation_genie_instance.all_dialogs):
                return self.analyze_query_no_dialog(query, dialog, debug, verbose, dialog_id, query_id)
            if dialog == True:
                if query_id is not None:
                    query_id = str(query_id)
                return self.analyze_query_attribute(query, dialog, debug, verbose, dialog_id, query_id)
            elif dialog == False or dialog is None:
                return self.analyze_query_no_dialog(query, dialog, debug, verbose, dialog_id, query_id)
            elif dialog == 'auto':
                inferred_dialog, confidence_level = self.autogenie_instance.analyze_followup_status(query, self.reserve_words, self.ignore_words)
                if inferred_dialog is True:
                    if query_id is not None:
                        query_id = str(query_id)
                    return self.analyze_query_attribute(query, inferred_dialog, debug, verbose, dialog_id, query_id, confidence_level)
                else:
                    return self.analyze_query_no_dialog(query, inferred_dialog, debug, verbose, dialog_id, query_id, confidence_level)
            else:
                raise RuntimeError("Expected values for \'dialog\' are True, False, or \"auto\"")
        elif self.processing_mode == 'gpt':
            if not bool(self.conversation_genie_instance.all_dialogs):
                return self.analyze_query_llm(query, dialog, dialog_id, query_id)
            if dialog == True:
                if query_id is not None:
                    query_id = str(query_id)
                return self.analyze_query_llm(query, dialog, dialog_id, query_id)
            elif dialog == False or dialog is None:
                return self.analyze_query_llm(query, dialog, dialog_id, query_id)
            elif dialog == 'auto':
                inferred_dialog, confidence_level = self.autogenie_instance.analyze_followup_status(query, self.reserve_words, self.ignore_words)
                if inferred_dialog is True:
                    if query_id is not None:
                        query_id = str(query_id)
                    return self.analyze_query_llm(query, inferred_dialog, debug, verbose, dialog_id, query_id, confidence_level)
                else:
                    return self.analyze_query_llm(query, inferred_dialog, debug, verbose, dialog_id, query_id, confidence_level)
            else:
                raise RuntimeError("Expected values for \'dialog\' are True, False, or \"auto\"")



    def analyze_query_no_dialog(self, query=None, dialog=None, debug=None, verbose=None, dialog_id=None, query_id=None, confidence_level = None):
        self.execution_durations = dict()
        self.dialog = dialog if dialog is not None else self.dialog
        self.debug = debug if debug is not None else self.debug
        self.verbose = verbose if verbose is not None else self.verbose

        # If not a follow-up query, reset the output variables.
        if not self.dialog:
            self.extracted_vis_type = None
            self.extracted_vis_token = None
            self.extracted_tasks = OrderedDict()
            self.extracted_attributes = OrderedDict()
            self.vis_list = None

        self.past_extracted_tasks = OrderedDict()
        self.past_extracted_attributes = OrderedDict()

        self.ambiguities= dict()
        self.ambiguities['attribute'] = dict()
        self.ambiguities['value'] = dict()


        # CLEAN AND PROCESS QUERY
        self.query_raw = query
        helpers.cond_print("Raw Query: " + self.query_raw, self.verbose)
        st = time.time()
        self.query_processed = self.query_genie_instance.process_query(self.query_raw)
        self.query_tokens = self.query_genie_instance.clean_query_and_get_query_tokens(self.query_processed, self.reserve_words, self.ignore_words)
        self.query_ngrams = self.query_genie_instance.get_query_ngrams(' '.join(self.query_tokens))
        helpers.cond_print("Processed Query: " + self.query_processed, self.verbose)
        self.execution_durations['clean_query'] = time.time() - st

        # DETECT EXPLICIT AND IMPLICIT ATTRIBUTES
        st = time.time()
        self.extracted_attributes = self.attribute_genie_instance.extract_attributes(self.query_ngrams)
        helpers.cond_print("Final Extracted Attributes: " + str(list(self.extracted_attributes.keys())), self.verbose)
        self.execution_durations['extract_attributes'] = time.time() - st
        self.query_for_task_inference = self.task_genie_instance.prepare_query_for_task_inference(self.query_processed)


        # DETECT EXPLICIT VISUALIZATION UTTERANCES
        st = time.time()
        self.extracted_vis_type, self.extracted_vis_token = self.vis_genie_instance.extract_vis_type(self.query_ngrams)
        self.execution_durations['extract_vis_type'] = time.time() - st

        # DETECT IMPLICIT AND EXPLICIT TASKS
        st = time.time()

        self.dependencies = self.task_genie_instance.create_dependency_tree(self.query_for_task_inference)
        task_map = self.task_genie_instance.extract_explicit_tasks_from_dependencies(self.dependencies)

        # Filters from Domain Values
        task_map = self.task_genie_instance.extract_explicit_tasks_from_domain_value(task_map)

        # At this stage, which attributes are encodeable?
        encodeable_attributes = self.attribute_genie_instance.get_encodeable_attributes()

        # INFER tasks based on (encodeable) attribute Datatypes
        task_map = self.task_genie_instance.extract_implicit_tasks_from_attributes(task_map, encodeable_attributes, self.dialog)

        # From the generated TaskMap, ensure that the task "keys" are NOT EMPTY LISTS
        self.extracted_tasks = self.task_genie_instance.filter_empty_tasks(task_map)

        self.execution_durations['extract_tasks'] = time.time() - st

        # RECOMMEND VISUALIZATIONS FROM ATTRIBUTES, TASKS, and VISUALIZATIONS
        st = time.time()
        if (bool(self.extracted_tasks) + bool(self.extracted_attributes) + bool(self.extracted_vis_type)) == 0:
            self.vis_list = list()
            output = {
            'status': 'SUCCESS' if len(self.vis_list) > 0 else 'FAILURE',
            'debug': {'execution_durations': self.execution_durations},
            'query_raw': self.query_raw,
            'query': self.query_processed,
            'dataset': self.data_url if self.data_url else self.data_value,
            'alias': self.alias_url if self.alias_url else self.alias_value,
            'visList': self.vis_list,
            'attributeMap': self.extracted_attributes,
            'taskMap': self.extracted_tasks,
            'followUpQuery': self.dialog,
            'contextObj': None,
            'attributeMapping': self.attribute_keyword_mapping,
            'followUpConfidence' : confidence_level,
            'ambiguity' : self.ambiguities
            }
            return output if self.debug else helpers.delete_keys_from_dict(output, keys=constants.keys_to_delete_in_output)

        # Final list of encodeable attributes in the VIS
        final_encodeable_attributes = self.attribute_genie_instance.update_encodeable_attributes_based_on_tasks()

        self.vis_list = self.vis_genie_instance.get_vis_list(attribute_list=final_encodeable_attributes)
        self.execution_durations['get_vis_list'] = time.time() - st
        self.execution_durations['total'] = sum(self.execution_durations.values())

        # Prepare output
        output = {
            'status': 'SUCCESS' if len(self.vis_list) > 0 else 'FAILURE',
            'debug': {'execution_durations': self.execution_durations},
            'query_raw': self.query_raw,
            'query': self.query_processed,
            'dataset': self.data_url if self.data_url else self.data_value,
            'alias': self.alias_url if self.alias_url else self.alias_value,
            'visList': self.vis_list,
            'attributeMap': self.extracted_attributes,
            'taskMap': self.extracted_tasks,
            'followUpQuery': self.dialog,
            'contextObj': None,
            'attributeMapping': self.attribute_keyword_mapping,
            'followUpConfidence' : confidence_level,
            'ambiguity' : self.ambiguities
        }
        if query_id is not None:
            query_id = str(query_id)
        if dialog_id is not None:
            dialog_id = str(dialog_id)
        return_convo, return_context = self.conversation_genie_instance.add_dialog(output, dialog_id, query_id)
        output['dialogId'] = str(return_convo)
        output['queryId'] = str(return_context)

        helpers.cond_print("Dialog ID: " + str(return_convo), self.verbose)
        helpers.cond_print("Query ID: " + str(return_context), self.verbose)

        return output if self.debug else helpers.delete_keys_from_dict(output, keys=constants.keys_to_delete_in_output)

    def analyze_query_llm(self, query=None, dialog=None, dialog_id=None, query_id=None):
        if dialog is None:
            query_prompt = self.promptgenie_instance.prompt
            query_prompt = re.sub(r"https://raw\.githubusercontent\.com\S*", self.data_url, query_prompt)
            dataset_sample = pd.read_csv(self.data_url, index_col = False)
            dataset_sample = dataset_sample.head(10).to_string(index=False)
            query_prompt = re.sub(r"<INSERT DATASET HERE>", dataset_sample, query_prompt)
            query_prompt = re.sub(r"<INSERT QUERY HERE>", query, query_prompt)
            message = self.chat_with_chatgpt(query_prompt)
            if query_id is not None:
                query_id = str(query_id)
            if dialog_id is not None:
                dialog_id = str(dialog_id)
            return_convo, return_context = self.conversation_genie_instance.add_dialog(message, dialog_id, query_id)
            message['dialogId'] = str(return_convo)
            message['queryId'] = str(return_context)
            return message
        elif dialog is True:
            context_obj = None
            if dialog_id == None and query_id == None:
                key_num = self.conversation_genie_instance.key_num - 1
                key_num = str(key_num)
                context_obj = self.conversation_genie_instance.all_dialogs[key_num][-1]
            elif dialog_id != None and query_id == None:
                dialog_id = str(dialog_id)
                context_obj = self.conversation_genie_instance.all_dialogs[dialog_id][-1]
            elif dialog_id == None and query_id != None:
                return None
            else:
                dialog_id = str(dialog_id)
                if not isinstance(query_id, str):
                    raise RuntimeError("Context id must be of type string")
                query_id = int(query_id)
                context_obj = self.conversation_genie_instance.all_dialogs[dialog_id][query_id]
            query_prompt = self.promptgenie_instance.prompt
            query_prompt = re.sub(r"https://raw\.githubusercontent\.com\S*", self.data_url, query_prompt)
            dataset_sample = pd.read_csv(self.data_url, index_col = False)
            dataset_sample = dataset_sample.head(10).to_string(index=False)
            query_prompt = re.sub(r"<INSERT DATASET HERE>", dataset_sample, query_prompt)
            query_prompt = re.sub(r"<INSERT QUERY HERE>", query, query_prompt)
            query_prompt = query_prompt + "\n" + "PREVIOUS ANALYTIC SPECIFICATION" + str(context_obj)
            message = self.chat_with_chatgpt(query_prompt)
            if query_id is not None:
                query_id = str(query_id)
            if dialog_id is not None:
                dialog_id = str(dialog_id)
            return_convo, return_context = self.conversation_genie_instance.add_dialog(message, dialog, dialog_id, query_id)
            message['dialogId'] = str(return_convo)
            message['queryId'] = str(return_context)
            return message

    def chat_with_chatgpt(self, new_prompt, model="gpt-4o-mini"):
        response = openai.ChatCompletion.create(
            model=model,
            messages=[{"role": "user",
               "content": new_prompt}],
            temperature=0
        )
        message = response.choices[0].message.content
        message = message.replace("\n", "")
        message = re.sub(r' +', ' ', message)
        # message = message.replace("json", "")
        message = message.replace("None", "\"None\"")
        message = message.replace("\"\"None\"\"", "\"None\"")
        message = message.replace("True", "true")
        message = message.replace("False", "false")
        if "json" == message[3:7]:
            message = message.replace("json", "", 1)
            message = message[3:]
            message = message[:-3]
        try:
            vlSpec = json.loads(message)
        except ValueError as e:
            return 'Invalid JSON'

        return vlSpec


    #put operation in meta
    #change operation to followup_type
    #put followup_type for cold queries, put it as None
    # meta: {followup_type: (add, remove, replace), followup_details : ({add : attr_value}, {remove: attr_value})}
    # use dependency parser to detect attributes now, and see if it's add remove or replace
    #create queries to see patterns for add, remove, and replace
    #also detect vis for attribute detection
    #concatenate group by and facet by groupby and facetby
    def analyze_query_attribute(self, query=None, dialog=None, debug=None, verbose=None, dialog_id=None, query_id=None, confidence_level=None):

        self.execution_durations = dict()
        self.dialog = dialog if dialog is not None else self.dialog
        self.debug = debug if debug is not None else self.debug
        self.verbose = verbose if verbose is not None else self.verbose

        context_obj = None
        if dialog is True:

            if dialog_id == None and query_id == None:
                key_num = self.conversation_genie_instance.key_num - 1
                key_num = str(key_num)

                context_obj = self.conversation_genie_instance.all_dialogs[key_num][-1]


            elif dialog_id != None and query_id == None:
                dialog_id = str(dialog_id)
                context_obj = self.conversation_genie_instance.all_dialogs[dialog_id][-1]
            elif dialog_id == None and query_id != None:
                return None
            else:
                dialog_id = str(dialog_id)
                if not isinstance(query_id, str):
                    raise RuntimeError("Context id must be of type string")
                query_id = int(query_id)
                context_obj = self.conversation_genie_instance.all_dialogs[dialog_id][query_id]
            helpers.cond_print("Previous Query: ", self.verbose)
            helpers.cond_print(context_obj['query_raw'], self.verbose)
            self.past_extracted_attributes = context_obj['attributeMap']
            self.past_ambiguities = context_obj['ambiguity']

            self.past_extracted_tasks = context_obj['taskMap']

            self.past_extracted_vis_type = context_obj['visList'][0]['visType']

            self.past_extracted_vis_list = context_obj['visList']
            self.past_attribute_keyword_mapping = context_obj['attributeMapping']

            if bool(self.past_ambiguities['attribute']):
                for attr_query in self.past_ambiguities['attribute']:
                    if not self.past_ambiguities['attribute'][attr_query]['selected'] == "NL4DV_Resolved":
                        self.updategenie_instance.auto_handle_attribute_ambiguity()
            if bool(self.past_ambiguities['value']):
                for attr_query in self.past_ambiguities['value']:
                    if not self.past_ambiguities['value'][attr_query]['selected'] == "NL4DV_Resolved":
                        raise Error("need to resolve ambiguity for value")

            self.ambiguities= dict()
            self.ambiguities['attribute'] = dict()
            self.ambiguities['value'] = dict()

        # CLEAN AND PROCESS QUERY
            self.query_raw = query
            helpers.cond_print("Raw Query: " + self.query_raw, self.verbose)
            st = time.time()
            self.query_processed = self.query_genie_instance.process_query(self.query_raw)


            self.query_tokens = self.query_genie_instance.clean_query_and_get_query_tokens(self.query_processed, self.reserve_words, self.ignore_words, dialog=self.dialog)
            self.query_ngrams = self.query_genie_instance.get_query_ngrams(' '.join(self.query_tokens))
            helpers.cond_print("Processed Query: " + self.query_processed, self.verbose)
            self.execution_durations['clean_query'] = time.time() - st

            # DETECT EXPLICIT AND IMPLICIT ATTRIBUTES
            st = time.time()
            self.extracted_attributes = self.attribute_genie_instance.extract_attributes(self.query_ngrams)

            helpers.cond_print("Final Extracted Attributes: " + str(list(self.extracted_attributes.keys())), self.verbose)
            self.execution_durations['extract_attributes'] = time.time() - st

            self.query_for_task_inference = self.task_genie_instance.prepare_query_for_task_inference(self.query_processed, dialog=True)
            self.dependencies = self.task_genie_instance.create_dependency_tree(self.query_for_task_inference)
            davo = self.conversation_genie_instance.get_followuptype_attribute_from_dependencies(self.dependencies)

            self.extracted_vis_type, self.extracted_vis_token = self.vis_genie_instance.extract_vis_type(self.query_ngrams, self.dependencies)
            if self.extracted_vis_type is None:
                self.extracted_vis_type = self.past_extracted_vis_type
            elif self.extracted_vis_type == "REMOVE":
                self.extracted_vis_type = None

            tasks_detected = False
            if bool(self.extracted_attributes):
                task_map = self.conversation_genie_instance.get_followuptype_task_from_dependencies(self.dependencies, self.query_ngrams)
                tasks_detected = True

            if davo[0] is None and len(self.extracted_attributes.keys()) != 0:
                davo[0] = "add"

            if davo[0] == "replace":
                if len(davo[1].values()) == 1:
                    if list(davo[1].values())[0] == "replace_to_add":
                        followup_dict = davo
                        removed_key, davo = self.conversation_genie_instance.implicit_replacement(followup_dict, self.past_extracted_attributes)
                        self.extracted_attributes[removed_key] = self.past_extracted_attributes[removed_key]

            if len(davo[1].keys()) != 0:
                for attr in self.extracted_attributes.keys():
                    if attr in davo[1].keys():
                        self.extracted_attributes[attr]['meta']['followup_type'] = davo[1][attr]
                    else:
                        self.extracted_attributes[attr]['meta']['followup_type'] = 'nothing'
                self.extracted_attributes, self.attribute_keyword_mapping = self.conversation_genie_instance.followup_conversation_attribute(self.extracted_attributes, self.past_extracted_attributes,
                    self.attribute_keyword_mapping, self.past_attribute_keyword_mapping, task_map)
            else:
                self.extracted_attributes, self.attribute_keyword_mapping = self.conversation_genie_instance.merge_update(
                    self.past_extracted_attributes, self.extracted_attributes, self.past_attribute_keyword_mapping, self.attribute_keyword_mapping)

            if not tasks_detected:
                task_map = self.conversation_genie_instance.get_followuptype_task_from_dependencies(self.dependencies, self.query_ngrams)

            if len(task_map.keys()) != 0:
                if 'filter' not in task_map.keys():
                    task_map = self.conversation_genie_instance.followup_conversation_task(task_map, self.past_extracted_tasks)
            task_map = self.task_genie_instance.extract_explicit_tasks_from_domain_value(task_map)

            if len(task_map.keys()) != 0 and 'filter' in task_map.keys():
                task_map = self.conversation_genie_instance.followup_filter_task(task_map, self.past_extracted_tasks, self.query_ngrams)

            if not bool(task_map):
                for i in self.past_extracted_tasks.keys():
                    if i in ['filter', 'derived_value', 'find_extremum', 'sort']:
                        task_map[i] = self.past_extracted_tasks[i]

            task_map = self.conversation_genie_instance.remove_old_tasks(task_map)

            # At this stage, which attributes are encodeable?
            encodeable_attributes = self.attribute_genie_instance.get_encodeable_attributes()

            # INFER tasks based on (encodeable) attribute Datatypes
            task_map = self.task_genie_instance.extract_implicit_tasks_from_attributes(task_map, encodeable_attributes, self.dialog)

            # From the generated TaskMap, ensure that the task "keys" are NOT EMPTY LISTS
            self.extracted_tasks = self.task_genie_instance.filter_empty_tasks(task_map)
            self.execution_durations['extract_tasks'] = time.time() - st

            # RECOMMEND VISUALIZATIONS FROM ATTRIBUTES, TASKS, and VISUALIZATIONS
            st = time.time()

            if (bool(self.extracted_tasks) + bool(self.extracted_attributes) + bool(self.extracted_vis_type)) == 0:
                self.vis_list = list()
                output = {
                'status': 'SUCCESS' if len(self.vis_list) > 0 else 'FAILURE',
                'debug': {'execution_durations': self.execution_durations},
                'query_raw': self.query_raw,
                'query': self.query_processed,
                'dataset': self.data_url if self.data_url else self.data_value,
                'alias': self.alias_url if self.alias_url else self.alias_value,
                'visList': self.vis_list,
                'attributeMap': self.extracted_attributes,
                'taskMap': self.extracted_tasks,
                'followUpQuery': self.dialog,
                'contextObj': None,
                'attributeMapping': self.attribute_keyword_mapping,
                'followUpConfidence' : confidence_level,
                'ambiguity' : self.ambiguities
                }
                return output if self.debug else helpers.delete_keys_from_dict(output, keys=constants.keys_to_delete_in_output)

            # Final list of encodeable attributes in the VIS
            final_encodeable_attributes = self.attribute_genie_instance.update_encodeable_attributes_based_on_tasks()

            self.vis_list = self.vis_genie_instance.get_vis_list(attribute_list=final_encodeable_attributes)

            self.execution_durations['get_vis_list'] = time.time() - st
            self.execution_durations['total'] = sum(self.execution_durations.values())
            current_ambiguity = dict()
            current_ambiguity['attribute'] = dict()
            current_ambiguity['value'] = dict()

            # Prepare output
            output = {
                'status': 'SUCCESS' if len(self.vis_list) > 0 else 'FAILURE',
                'debug': {'execution_durations': self.execution_durations},
                'query_raw': self.query_raw,
                'query': self.query_processed,
                'dataset': self.data_url if self.data_url else self.data_value,
                'alias': self.alias_url if self.alias_url else self.alias_value,
                'visList': self.vis_list,
                'attributeMap': self.extracted_attributes,
                'taskMap': self.extracted_tasks,
                'followUpQuery': self.dialog,
                'contextObj': None,
                'attributeMapping': self.attribute_keyword_mapping,
                'followUpConfidence': confidence_level,
                'ambiguity' : self.ambiguities
            }

            if query_id is not None:
                query_id = str(query_id)
            if dialog_id is not None:
                dialog_id = str(dialog_id)
            return_conversation, return_context = self.conversation_genie_instance.add_dialog(output, dialog, dialog_id, query_id)
            if len(self.ambiguities['attribute']) != 0 or len(self.ambiguities['value']) != 0:
                for key in self.ambiguities['attribute']:
                    if key not in self.past_ambiguities['attribute']:
                        current_ambiguity['attribute'][key] = self.ambiguities['attribute'][key]
                    else:
                        current_ambiguity['attribute'][key] = self.past_ambiguities['attribute'][key]
                for key in self.ambiguities['value']:
                    if key not in self.past_ambiguities['value']:
                        current_ambiguity['value'][key] = self.ambiguities['value'][key]
                    else:
                        current_ambiguity['value'][key] = self.past_ambiguities['value'][key]
                self.ambiguities = current_ambiguity
                output = self.updategenie_instance.update_vis(str(return_conversation), str(return_context))

            output['dialogId'] = str(return_conversation)
            output['queryId'] = str(return_context)

            helpers.cond_print("Dialog ID: " + str(return_conversation), self.verbose)
            helpers.cond_print("Query ID: " + str(return_context), self.verbose)

            return output if self.debug else helpers.delete_keys_from_dict(output, keys=constants.keys_to_delete_in_output)

    def update_query(self, ambiguity_obj):
        if 'query_id' in ambiguity_obj.keys() and 'dialog_id' in ambiguity_obj.keys():
            query_id = ambiguity_obj['query_id']
            dialog_id = ambiguity_obj['dialog_id']
        else:
            dialog_id = self.conversation_genie_instance.most_recent_dialog_id[-1]
            query_id = self.conversation_genie_instance.most_recent_query_id[-1]
        if "attribute" in ambiguity_obj.keys():
            for key in ambiguity_obj['attribute']:
                self.ambiguities['attribute'][key]['selected'] = ambiguity_obj['attribute'][key]
        if "value" in ambiguity_obj.keys():
            for key in ambiguity_obj['value']:
                self.ambiguities['value'][key]['selected'] = ambiguity_obj['value'][key]
        return self.updategenie_instance.update_vis(dialog_id, query_id)

    # Update the attribute datatypes that were not correctly detected by NL4DV
    def set_attribute_datatype(self, attr_type_obj):
        return self.data_genie_instance.set_attribute_datatype(attr_type_obj=attr_type_obj)

    # Set Label attribute for the dataset, i.e. one that defines what the dataset is about.
    # e.g. "Correlate horsepower and MPG for sports car models" should NOT apply an explicit attribute for models since there are two explicit attributes already present.
    def set_label_attribute(self, label_attribute):
        return self.data_genie_instance.set_label_attribute(label_attribute=label_attribute)

    # WORDS that should be IGNORED in the query, i.e. NOT lead to the detection of attributes and tasks
    # `Movie` in movies dataset
    # `Car` in cars dataset
    def set_ignore_words(self, ignore_words):
        return self.data_genie_instance.set_ignore_words(ignore_words=ignore_words)

    # Custom STOPWORDS that should NOT removed from the query, as they might be present in the domain.
    # e.g. `A` in grades dataset
    def set_reserve_words(self, reserve_words):
        return self.data_genie_instance.set_reserve_words(reserve_words=reserve_words)

    # Sets the AliasMap
    def set_alias_map(self, alias_value=None, alias_url=None):
        return self.data_genie_instance.set_alias_map(alias_value=alias_value, alias_url=alias_url)

    def set_implicit_followup_keywords(self, implicit_followup_keywords=None):
        if implicit_followup_keywords is not None:
            self.implicit_followup_keyword_map = implicit_followup_keywords

    def set_explicit_followup_keywords(self, explicit_followup_keywords=None):
        if explicit_followup_keywords is not None:
            self.followup_keyword_map = explicit_followup_keywords

    # Sets the Dataset
    def set_data(self, data_url=None, data_value=None):
        return self.data_genie_instance.set_data(data_url=data_url, data_value=data_value)

    # Sets the String Matching, Domain Word Limit, ... Thresholds
    def set_thresholds(self, thresholds):
        for t in thresholds:
            if t in self.match_thresholds and (isinstance(thresholds[t], float) or isinstance(thresholds[t], int)):
                self.match_thresholds[t] = thresholds[t]

    # Sets the Scoring Weights for the way attributes / tasks and visualizations are detected.
    def set_importance_scores(self, scores):
        # domain = {'attribute', 'task', 'vis'}
        for domain in scores.keys():
            if domain in self.match_scores and isinstance(scores[domain], dict):
                # setting = {'attribute_exact_match', 'attribute_similarity_match', 'attribute_alias_exact_match', ... so on}
                for setting in scores[domain].keys():
                    if setting in self.match_scores[domain] and isinstance(scores[domain][setting], float):
                        self.match_scores[domain][setting] = scores[domain][setting]

    # Flush specific conversation and context
    def delete_dialogs(self, dialog_id=None, query_id = None):
        return self.conversation_genie_instance.delete_dialogs(dialog_id=dialog_id, query_id=query_id)

    def get_dialogs(self, dialog_id, query_id):
        return self.conversation_genie_instance.get_dialogs(dialog_id=dialog_id, query_id=query_id)

    def undo(self):
        return self.conversation_genie_instance.undo()

    # Get the dataset metadata
    def get_metadata(self):
        return json.loads(json.dumps(self.data_genie_instance.data_attribute_map, cls=helpers.CustomJSONEncoder))

    # Create a dependency parser instance
    def set_dependency_parser(self, config):
        if isinstance(config, dict):
            helpers.cond_print("Dependency Parser: " + config["name"], self.verbose)
            self.dependency_parser = config["name"]
            if config["name"] == "spacy":
                """
                    Sets the model and returns the Spacy NLP instance. Example ways from the Spacy docs:
                    spacy.load("en") # shortcut link
                    spacy.load("en_core_web_sm") # package
                    spacy.load("/path/to/en") # unicode path
                    spacy.load(Path("/path/to/en")) # pathlib Path
                """
                self.dependency_parser_instance = spacy.load(config["model"])

            elif config["name"] == "corenlp":
                if 'CLASSPATH' not in os.environ:
                    os.environ['CLASSPATH'] = ""

                cpath = config["model"] + os.pathsep + config["parser"]
                if cpath not in os.environ['CLASSPATH']:
                    os.environ['CLASSPATH'] = cpath + os.pathsep + os.environ['CLASSPATH']

                # TODO:- DEPRECATED
                self.dependency_parser_instance = StanfordDependencyParser(path_to_models_jar=config["model"],
                                                                           encoding='utf8')
            elif config["name"] == "corenlp-server":
                # Requires the CoreNLPServer running in the background at the below URL (generally https://localhost:9000)
                # Start server by running the following command in the JARs directory.
                # `java -mx4g -cp "*" edu.stanford.nlp.pipeline.StanfordCoreNLPServer -annotators "tokenize,ssplit,pos,lemma,parse,sentiment" -port 9000 -timeout 30000`
                self.dependency_parser_instance = CoreNLPDependencyParser(url=config["url"])
